import pandas as pd


def _assess_pandas_object_column(pandas_df: pd.DataFrame, col: str) -> tuple[str, str]:
    """tries to determine less generic type for object columns

    Parameters
    ----------
    pandas_df: pd.DataFrame
        pandas df to assess

    col: str
        column to assess

    Returns
    ----------
    pandas_col_type: str
        deduced pandas col type

    polars_col_type: str
        deduced polars col type
    """

    if pandas_df[col].dtype.name != "object":
        msg = "_assess_pandas_object_column only works with object dtype columns"
        raise TypeError(
            msg,
        )

    pandas_col_type = "object"
    polars_col_type = "Object"

    # pandas would assign dtype object to bools with nulls, but have values like True
    # it would also assign all null cols to object, but have values like None
    # creating a polars col with object type would give values like 'true', 'none'
    # overwrite these cases for better handling
    if pandas_df[col].notna().sum() == 0:
        pandas_col_type = "null"
        polars_col_type = "Unknown"

    # check if all non-null values are bool
    elif sum(isinstance(value, bool) for value in pandas_df[col] if value):
        pandas_col_type = "bool"
        polars_col_type = "Boolean"

    # we may wish to add more cases in future, but starting with these

    return pandas_col_type, polars_col_type



import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.numeric.test_BaseNumericTransformer import (
    BaseNumericTransformerInitTests,
    BaseNumericTransformerTransformTests,
)
from tubular.numeric import CutTransformer


class TestInit(BaseNumericTransformerInitTests):
    """Tests for CutTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CutTransformer"

    def test_new_column_name_type_error(self):
        """Test that an exception is raised if new_column_name is not a str."""
        with pytest.raises(
            TypeError,
            match="CutTransformer: new_column_name must be a str",
        ):
            CutTransformer(column="b", new_column_name=1)

    def test_cut_kwargs_type_error(self):
        """Test that an exception is raised if cut_kwargs is not a dict."""
        with pytest.raises(
            TypeError,
            match=r"""cut_kwargs should be a dict but got type \<class 'int'\>""",
        ):
            CutTransformer(column="b", new_column_name="a", cut_kwargs=1)

    def test_cut_kwargs_key_type_error(self):
        """Test that an exception is raised if cut_kwargs has keys which are not str."""
        with pytest.raises(
            TypeError,
            match=r"""CutTransformer: unexpected type \(\<class 'int'\>\) for cut_kwargs key in position 1, must be str""",
        ):
            CutTransformer(
                new_column_name="a",
                column="b",
                cut_kwargs={"a": 1, 2: "b"},
            )


class TestTransform(BaseNumericTransformerTransformTests):
    """Tests for CutTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CutTransformer"

    def expected_df_1():
        """Expected output for test_expected_output."""
        df = d.create_df_9()

        df["d"] = pd.Categorical(
            values=["c", "b", "a", "d", "e", "f"],
            categories=["a", "b", "c", "d", "e", "f"],
            ordered=True,
        )

        return df

    @pytest.mark.parametrize(
        ("df_generator"),
        [
            d.create_df_2,  # int
            d.create_bool_and_float_df,  # float
        ],
    )
    def test_numeric_passes(self, initialized_transformers, df_generator):
        """Test check passes if self.columns numeric in X - this transformer does not work on all null column
        so overload this test"""
        df = df_generator()
        # add in 'target column' for and additional numeric column fit
        df["c"] = [1] * len(df)
        df["b"] = [1] * len(df)

        x = initialized_transformers[self.transformer_name]
        x.columns = ["a", "b"]

        numeric_df = pd.DataFrame({col: df["c"] for col in [*x.columns, "c"]})
        x.fit(numeric_df, numeric_df["c"])

        x.transform(df)

    def test_output_from_cut_assigned_to_column(self, mocker):
        """Test that the output from pd.cut is assigned to column with name new_column_name."""
        df = d.create_df_9()

        x = CutTransformer(column="c", new_column_name="c_new", cut_kwargs={"bins": 2})

        cut_output = [1, 2, 3, 4, 5, 6]

        mocker.patch("pandas.cut", return_value=cut_output)

        df_transformed = x.transform(df)

        assert (
            df_transformed["c_new"].tolist() == cut_output
        ), "unexpected values assigned to c_new column"

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_9(), expected_df_1()),
    )
    def test_expected_output(self, df, expected):
        """Test input data is transformed as expected."""
        cut_1 = CutTransformer(
            column="c",
            new_column_name="d",
            cut_kwargs={
                "bins": [0, 1, 2, 3, 4, 5, 6],
                "labels": ["a", "b", "c", "d", "e", "f"],
            },
        )

        df_transformed = cut_1.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="CutTransformer.transform output",
        )



import datetime
import re

import narwhals as nw
import pandas as pd
import polars as pl
import pytest

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
)
from tests.dates.test_BaseGenericDateTransformer import (
    GenericDatesMixinTransformTests,
    create_date_diff_different_dtypes,
)
from tests.utils import assert_frame_equal_dispatch
from tubular.dates import BetweenDatesTransformer


class TestInit(
    ColumnStrListInitTests,
    NewColumnNameInitMixintests,
    DropOriginalInitMixinTests,
):
    "tests for BetweenDatesTransformer.__init__."

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BetweenDatesTransformer"

    @pytest.mark.parametrize(
        ("param", "value"),
        [
            ("upper_inclusive", 1000),
            ("lower_inclusive", "hi"),
        ],
    )
    def test_inclusive_args_non_bool_error(self, param, value):
        """Test that an exception is raised if upper_inclusive not a bool."""

        param_dict = {param: value}
        with pytest.raises(
            TypeError,
            match=f"BetweenDatesTransformer: {param} should be a bool",
        ):
            BetweenDatesTransformer(
                columns=["a", "b", "c"],
                new_column_name="d",
                **param_dict,
            )

    @pytest.mark.parametrize(
        "columns",
        [
            ["a", "b"],
            ["a", "b", "c", "d"],
        ],
    )
    def test_wrong_col_count_error(self, columns):
        """Test that an exception is raised if too many/too few columns."""

        with pytest.raises(
            ValueError,
            match="BetweenDatesTransformer: This transformer works with three columns only",
        ):
            BetweenDatesTransformer(
                columns=columns,
                new_column_name="d",
            )


def expected_df_1(library="pandas"):
    """Expected output from transform in test_output."""
    df = d.create_is_between_dates_df_1(library=library)

    df = nw.from_native(df)
    native_backend = nw.get_native_namespace(df)
    df = df.with_columns(
        nw.new_series(
            name="d",
            values=[True, False],
            backend=native_backend.__name__,
        ),
    )

    return df.to_native()


def expected_df_2(library="pandas"):
    """Expected output from transform in test_output_both_exclusive."""
    df = d.create_is_between_dates_df_2(library=library)

    df = nw.from_native(df)
    native_backend = nw.get_native_namespace(df)
    df = df.with_columns(
        nw.new_series(
            name="e",
            values=[False, False, True, True, False, False],
            backend=native_backend.__name__,
        ),
    )

    return df.to_native()


def expected_df_3(library="pandas"):
    """Expected output from transform in test_output_lower_exclusive."""
    df = d.create_is_between_dates_df_2(library=library)

    df = nw.from_native(df)
    native_backend = nw.get_native_namespace(df)
    df = df.with_columns(
        nw.new_series(
            name="e",
            values=[False, False, True, True, True, False],
            backend=native_backend.__name__,
        ),
    )

    return df.to_native()


def expected_df_4(library="pandas"):
    """Expected output from transform in test_output_upper_exclusive."""
    df = d.create_is_between_dates_df_2(library=library)

    df = nw.from_native(df)
    native_backend = nw.get_native_namespace(df)
    df = df.with_columns(
        nw.new_series(
            name="e",
            values=[False, True, True, True, False, False],
            backend=native_backend.__name__,
        ),
    )

    return df.to_native()


def expected_df_5(library="pandas"):
    """Expected output from transform in test_output_both_inclusive."""
    df = d.create_is_between_dates_df_2(library=library)

    df = nw.from_native(df)
    native_backend = nw.get_native_namespace(df)
    df = df.with_columns(
        nw.new_series(
            name="e",
            values=[False, True, True, True, True, False],
            backend=native_backend.__name__,
        ),
    )

    return df.to_native()


class TestTransform(
    GenericTransformTests,
    GenericDatesMixinTransformTests,
    DropOriginalTransformMixinTests,
):
    """Tests for BetweenDatesTransformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BetweenDatesTransformer"

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_is_between_dates_df_1(library="pandas"),
                expected_df_1(library="pandas"),
            ),
            (
                d.create_is_between_dates_df_1(library="polars"),
                expected_df_1(library="polars"),
            ),
        ],
    )
    def test_output(self, df, expected):
        """Test the output of transform is as expected."""
        x = BetweenDatesTransformer(
            columns=["a", "b", "c"],
            new_column_name="d",
            lower_inclusive=False,
            upper_inclusive=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_is_between_dates_df_2(library="pandas"),
                expected_df_2(library="pandas"),
            ),
            (
                d.create_is_between_dates_df_2(library="polars"),
                expected_df_2(library="polars"),
            ),
        ],
    )
    def test_output_both_exclusive(self, df, expected):
        """Test the output of transform is as expected if both limits are exclusive."""
        x = BetweenDatesTransformer(
            columns=["a", "b", "c"],
            new_column_name="e",
            lower_inclusive=False,
            upper_inclusive=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_is_between_dates_df_2(library="pandas"),
                expected_df_3(library="pandas"),
            ),
            (
                d.create_is_between_dates_df_2(library="polars"),
                expected_df_3(library="polars"),
            ),
        ],
    )
    def test_output_lower_exclusive(self, df, expected):
        """Test the output of transform is as expected if the lower limits are exclusive only."""
        x = BetweenDatesTransformer(
            columns=["a", "b", "c"],
            new_column_name="e",
            lower_inclusive=False,
            upper_inclusive=True,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_is_between_dates_df_2(library="pandas"),
                expected_df_4(library="pandas"),
            ),
            (
                d.create_is_between_dates_df_2(library="polars"),
                expected_df_4(library="polars"),
            ),
        ],
    )
    def test_output_upper_exclusive(self, df, expected):
        """Test the output of transform is as expected if the upper limits are exclusive only."""
        x = BetweenDatesTransformer(
            columns=["a", "b", "c"],
            new_column_name="e",
            lower_inclusive=True,
            upper_inclusive=False,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    @pytest.mark.parametrize(
        ("df", "expected"),
        [
            (
                d.create_is_between_dates_df_2(library="pandas"),
                expected_df_5(library="pandas"),
            ),
            (
                d.create_is_between_dates_df_2(library="polars"),
                expected_df_5(library="polars"),
            ),
        ],
    )
    def test_output_both_inclusive(self, df, expected):
        """Test the output of transform is as expected if the both limits are inclusive."""
        x = BetweenDatesTransformer(
            columns=["a", "b", "c"],
            new_column_name="e",
            lower_inclusive=True,
            upper_inclusive=True,
        )

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(expected, df_transformed)

    def test_warning_message(self):
        """Test a warning is generated if not all the values in column_upper are greater than or equal to column_lower."""
        x = BetweenDatesTransformer(
            columns=["a", "b", "c"],
            new_column_name="e",
            lower_inclusive=True,
            upper_inclusive=True,
        )

        df = d.create_is_between_dates_df_2()
        df = nw.from_native(df)

        df = (
            df.with_row_index("i")
            .with_columns(
                c=nw.when(nw.col("i") == 0)
                .then(datetime.datetime(1989, 3, 1, tzinfo=datetime.timezone.utc))
                .otherwise("c"),
            )
            .drop("i")
        ).to_native()

        with pytest.warns(Warning, match="not all c are greater than or equal to a"):
            x.transform(df)

    @pytest.mark.parametrize(
        ("columns"),
        [
            ["a_date", "b_date", "c_date"],
            ["a_date", "b_date", "c_datetime"],
            ["a_date", "b_datetime", "c_datetime"],
            ["a_datetime", "b_date", "c_date"],
            ["a_datetime", "b_date", "c_datetime"],
            ["a_datetime", "b_datetime", "c_date"],
        ],
    )
    @pytest.mark.parametrize(
        ("library"),
        ["pandas", "polars"],
    )
    def test_output_different_date_dtypes(self, columns, library):
        """Test the output of transform is as expected if both limits are exclusive."""
        x = BetweenDatesTransformer(
            columns=columns,
            new_column_name="e",
            lower_inclusive=False,
            upper_inclusive=False,
        )

        df = d.create_is_between_dates_df_3(library=library)
        output = [False, False, True, True, False, False]
        df = nw.from_native(df)
        expected = df.clone()
        if library == "pandas":
            expected = expected.with_columns(
                [
                    nw.from_native(pd.Series(output), series_only=True).alias("e"),
                ],
            ).to_native()
        if library == "polars":
            expected = expected.with_columns(
                [
                    nw.from_native(pl.Series(output), series_only=True).alias("e"),
                ],
            ).to_native()

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

    # overloading below test as column count is different for this one
    @pytest.mark.parametrize(
        ("columns, datetime_col, date_col"),
        [
            (["date_col_1", "datetime_col_2", "date_col_2"], 1, 0),
            (["datetime_col_1", "date_col_2", "datetime_col_2"], 0, 1),
        ],
    )
    @pytest.mark.parametrize(
        ("library"),
        ["pandas", "polars"],
    )
    def test_mismatched_datetypes_error(
        self,
        columns,
        datetime_col,
        date_col,
        uninitialized_transformers,
        library,
    ):
        "Test that transform raises an error if one column is a date and one is datetime"

        x = uninitialized_transformers[self.transformer_name](
            columns=columns,
            new_column_name="c",
        )

        df = create_date_diff_different_dtypes(library=library)

        present_types = (
            {"datetime64", "date32[pyarrow]"}
            if datetime_col == 0
            else {"date32[pyarrow]", "datetime64"}
        )
        msg = re.escape(
            f"Columns fed to datetime transformers should be ['datetime64', 'date32[pyarrow]'] and have consistent types, but found {present_types}. Please use ToDatetimeTransformer to standardise",
        )
        with pytest.raises(
            TypeError,
            match=msg,
        ):
            x.transform(df)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BetweenDatesTransformer"



import numpy as np
import pandas as pd
import pytest
import test_aide as ta
from test_BaseNominalTransformer import GenericNominalTransformTests

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    OtherBaseBehaviourTests,
)
from tubular.nominal import NominalToIntegerTransformer


class TestInit(ColumnStrListInitTests):
    """Tests for NominalToIntegerTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NominalToIntegerTransformer"

    def test_start_encoding_not_int_error(self):
        """Test that an exception is raised if start_encoding is not an int."""
        with pytest.raises(ValueError):
            NominalToIntegerTransformer(columns="a", start_encoding="a")


class TestFit(GenericFitTests):
    """Tests for NominalToIntegerTransformer.fit()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NominalToIntegerTransformer"

    def test_learnt_values(self):
        """Test that the impute values learnt during fit are expected."""
        df = d.create_df_1()

        x = NominalToIntegerTransformer(columns=["a", "b"], start_encoding=1)

        x.fit(df)

        ta.classes.test_object_attributes(
            obj=x,
            expected_attributes={
                "mappings": {
                    "a": {k: i for i, k in enumerate(df["a"].unique(), 1)},
                    "b": {k: i for i, k in enumerate(df["b"].unique(), 1)},
                },
            },
            msg="mappings attribute",
        )

    def test_error_for_too_many_levels(self):
        "test that transformer.transform errors for column with too many levels"
        transformer = NominalToIntegerTransformer(columns=["a"])

        df = pd.DataFrame(
            {
                "a": list(range(1000)),
                "b": list(range(1000)),
            },
        )

        with pytest.raises(
            ValueError,
            match="NominalToIntegerTransformer: column a has too many levels to encode",
        ):
            transformer.fit(df, df["b"])


class TestTransform(GenericNominalTransformTests):
    """Tests for NominalToIntegerTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NominalToIntegerTransformer"

    def expected_df_1():
        """Expected output for test_expected_output."""
        df = pd.DataFrame(
            {"a": [1, 2, 3, 4, 5, 6], "b": ["a", "b", "c", "d", "e", "f"]},
        )

        df["a"] = (
            df["a"]
            .replace({k: i for i, k in enumerate(df["a"].unique())})
            .astype(np.int8)
        )

        df["b"] = (
            df["b"]
            .replace({k: i for i, k in enumerate(df["b"].unique())})
            .astype(np.int8)
        )

        return df

    def test_learnt_values_not_modified(self):
        """Test that the mappings from fit are not changed in transform."""
        df = d.create_df_1()

        x = NominalToIntegerTransformer(columns=["a", "b"])

        x.fit(df)

        x2 = NominalToIntegerTransformer(columns=["a", "b"])

        x2.fit_transform(df)

        ta.equality.assert_equal_dispatch(
            expected=x.mappings,
            actual=x2.mappings,
            msg="Impute values not changed in transform",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_1(), expected_df_1()),
    )
    def test_expected_output(self, df, expected):
        """Test that the output is expected from transform."""
        x = NominalToIntegerTransformer(columns=["a", "b"])

        # set the mapping dict directly rather than fitting x on df so test works with helpers
        x.mappings = {
            "a": {1: 0, 2: 1, 3: 2, 4: 3, 5: 4, 6: 5},
            "b": {"a": 0, "b": 1, "c": 2, "d": 3, "e": 4, "f": 5},
        }

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="Unexpected values in NominalToIntegerTransformer.transform",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NominalToIntegerTransformer"



import pandas as pd
import polars as pl
from narwhals.typing import FrameT
from pandas.testing import assert_frame_equal as assert_pandas_frame_equal
from polars.testing import assert_frame_equal as assert_polars_frame_equal

from tubular._utils import (
    _assess_pandas_object_column,  # noqa: PLC2701, purposefully using private method in tests
)

PANDAS_TO_POLARS_TYPES = {
    "int64": pl.Int64,
    "int32": pl.Int32,
    "int16": pl.Int16,
    "int8": pl.Int8,
    "float64": pl.Float64,
    "float32": pl.Float32,
    "object": pl.Utf8,
    "str": pl.String,
    "bool": pl.Boolean,
    "datetime64[ns]": pl.Datetime,
    "datetime64[ns, UTC]": pl.Datetime(time_zone="UTC"),
    "date32[day][pyarrow]": pl.Date,
    # this is not a pandas type, but include to help manage null column handling
    "null": pl.Unknown,
}


def _align_pandas_and_polars_dtypes(
    pandas_df: pd.DataFrame,
    polars_df: pl.DataFrame,
) -> pl.DataFrame:
    """helper to ensure that dtypes are aligned between polars/pandas dfs used in tests. Avoids e.g.
    [0, None] being inferred as float in pandas and int in polars and complicating tests

    Parameters
    ----------
    pandas_df : pd.DataFrame
        pandas df to extract types from

    polars_df : pl.DataFrame
        polars df to apply types to

    Returns
    ----------

    pl.DataFrame: polars df with types converted to align with pandas_df

    """

    for col in pandas_df:
        pandas_col_type = str(pandas_df[col].dtype)

        # pandas would assign dtype object to bools with nulls, but have values like True
        # it would also assign all null cols to object, but have values like None
        # creating a polars col with object type would give values like 'true', 'none'
        # overwrite these cases for better handling
        if pandas_col_type == "object":
            pandas_col_type, polars_col_type = _assess_pandas_object_column(
                pandas_df=pandas_df,
                col=col,
            )

        polars_col_type = PANDAS_TO_POLARS_TYPES[pandas_col_type]
        polars_df = polars_df.with_columns(polars_df[col].cast(polars_col_type))

    return polars_df


def assert_frame_equal_dispatch(df1: FrameT, df2: FrameT) -> None:
    """fixture to return correct pandas/polars assert_frame_equal method

    Parameters
    ----------
    df1 : pd.DataFrame
        first df for comparison

    df1 : pd.DataFrame
        second df for comparison

    """

    if isinstance(df1, pd.DataFrame) and isinstance(df2, pd.DataFrame):
        return assert_pandas_frame_equal(df1, df2)

    if isinstance(df1, pl.DataFrame) and isinstance(df2, pl.DataFrame):
        return assert_polars_frame_equal(df1, df2)

    invalid_request_error = "tubular is setup to handle only pandas or polars inputs, and dfs input to this function should be from same library"
    raise ValueError(invalid_request_error)


def dataframe_init_dispatch(
    dataframe_dict: dict,
    library: str,
) -> FrameT:
    """
    Initialize a DataFrame using either Pandas or Polars library based on the specified library name.

    Parameters:
    dataframe_dict (dict): A dictionary where keys are column names and values are lists of column data.
    library (str): The name of the library to use for DataFrame creation. Should be either "pandas" or "polars".

    Returns:
    pl.DataFrame | pd.DataFrame: A DataFrame object from the specified library.

    Raises:
    ValueError: If the `library` parameter is not "pandas" or "polars".
    """

    pandas_df = pd.DataFrame(dataframe_dict)

    # narwhals only supports pyarrow date type, so convert
    for col in pandas_df:
        if pd.api.types.infer_dtype(pandas_df[col]) == "date":
            pandas_df[col] = pandas_df[col].astype("date32[pyarrow]")

        # ensure datetimes setup as expected
        if pd.api.types.is_datetime64_any_dtype(pandas_df[col]):
            pandas_df[col] = pd.to_datetime(pandas_df[col])

    if library == "pandas":
        return pandas_df

    if library == "polars":
        polars_df = pl.DataFrame(dataframe_dict)

        return _align_pandas_and_polars_dtypes(pandas_df, polars_df)

    library_error_message = (
        "The library parameter should be either 'pandas' or 'polars'."
    )
    raise ValueError(library_error_message)



"""This module contains transformers that apply encodings to nominal columns."""
from __future__ import annotations

import copy
import warnings
from typing import TYPE_CHECKING, Literal, Optional, Union

import narwhals as nw
import narwhals.selectors as ncs
import numpy as np
from beartype import beartype

from tubular.base import BaseTransformer
from tubular.imputers import MeanImputer, MedianImputer
from tubular.mapping import BaseMappingTransformMixin
from tubular.mixins import DropOriginalMixin, SeparatorColumnMixin, WeightColumnMixin

if TYPE_CHECKING:
    import pandas as pd
    from narwhals.typing import FrameT


class BaseNominalTransformer(BaseTransformer):
    """
    Base Transformer extension for nominal transformers.

    Attributes
    ----------

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    FITS = False

    @nw.narwhalify
    def check_mappable_rows(self, X: FrameT) -> None:
        """Method to check that all the rows to apply the transformer to are able to be
        mapped according to the values in the mappings dict.

        Raises
        ------
        ValueError
            If any of the rows in a column (c) to be mapped, could not be mapped according to
            the mapping dict in mappings[c].

        """
        self.check_is_fitted(["mappings"])

        for c in self.columns:
            mappable_rows = X.select(
                nw.col(c).is_in(list(self.mappings[c])).sum(),
            ).item()

            if mappable_rows < X.shape[0]:
                msg = f"{self.classname()}: nulls would be introduced into column {c} from levels not present in mapping"
                raise ValueError(msg)

    @nw.narwhalify
    def transform(self, X: FrameT) -> None:
        """Base nominal transformer transform method.  Checks that all the rows are able to be
        mapped according to the values in the mappings dict and calls the BaseTransformer transform method.

        Parameters
        ----------
        X : FrameT
            Data to apply nominal transformations to.

        Returns
        -------
        X : FrameT
            Input X.

        """

        # specify which class to prevent additional inheritance calls
        X = BaseTransformer.transform(self, X)

        self.check_mappable_rows(X)

        return X


class NominalToIntegerTransformer(BaseNominalTransformer, BaseMappingTransformMixin):
    """Transformer to convert columns containing nominal values into integer values.

    The nominal levels that are mapped to integers are not ordered in any way.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    start_encoding : int, default = 0
        Value to start the encoding from e.g. if start_encoding = 0 then the encoding would be
        {'A': 0, 'B': 1, 'C': 3} etc.. or if start_encoding = 5 then the same encoding would be
        {'A': 5, 'B': 6, 'C': 7}. Can be positive or negative.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    start_encoding : int
        Value to start the encoding / mapping of nominal to integer from.

    mappings : dict
        Created in fit. A dict of key (column names) value (mappings between levels and integers for given
        column) pairs.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    FITS = True

    def __init__(
        self,
        columns: str | list[str] | None = None,
        start_encoding: int = 0,
        **kwargs: dict[str, bool],
    ) -> None:
        BaseNominalTransformer.__init__(self, columns=columns, **kwargs)

        # this transformer shouldn't really be used with huge numbers of levels
        # so setup to use int8 type
        # if there are more levels than this, will get a type error
        self.return_dtypes = {c: "Int8" for c in self.columns}

        if not isinstance(start_encoding, int):
            msg = f"{self.classname()}: start_encoding should be an integer"
            raise ValueError(msg)

        self.start_encoding = start_encoding

    def fit(self, X: pd.DataFrame, y: pd.Series | None = None) -> pd.DataFrame:
        """Creates mapping between nominal levels and integer values for categorical variables.

        Parameters
        ----------
        X : pd.DataFrame
            Data to fit the transformer on, this sets the nominal levels that can be mapped.

        y : None or pd.DataFrame or pd.Series, default = None
            Optional argument only required for the transformer to work with sklearn pipelines.

        """
        BaseNominalTransformer.fit(self, X, y)

        self.mappings = {}

        for c in self.columns:
            col_values = X[c].unique()

            self.mappings[c] = {
                k: i for i, k in enumerate(col_values, self.start_encoding)
            }

            # if more levels than int8 type can handle, then error
            if len(self.mappings[c]) > 127:
                msg = f"{self.classname()}: column {c} has too many levels to encode"
                raise ValueError(
                    msg,
                )

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform method to apply integer encoding stored in the mappings attribute to
        each column in the columns attribute.

        This method calls the check_mappable_rows method from BaseNominalTransformer to check that
        all rows can be mapped then transform from BaseMappingTransformMixin to apply the
        standard pd.Series.map method.

        Parameters
        ----------
        X : pd.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with levels mapped according to mappings dict.

        """

        X = super().transform(X)

        return BaseMappingTransformMixin.transform(self, X)


class GroupRareLevelsTransformer(BaseTransformer, WeightColumnMixin):

    """Transformer to group together rare levels of nominal variables into a new level,
    labelled 'rare' (by default).

    Rare levels are defined by a cut off percentage, which can either be based on the
    number of rows or sum of weights. Any levels below this cut off value will be
    grouped into the rare level.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    cut_off_percent : float, default = 0.01
        Cut off for the percent of rows or percent of weight for a level, levels below
        this value will be grouped.

    weights_column : None or str, default = None
        Name of weights column that should be used so cut_off_percent applies to sum of weights
        rather than number of rows.

    rare_level_name : any,default = 'rare'.
        Must be of the same type as columns.
        Label for the new 'rare' level.

    record_rare_levels : bool, default = False
        If True, an attribute called rare_levels_record_ will be added to the object. This will be a dict
        of key (column name) value (level from column considered rare according to cut_off_percent) pairs.
        Care should be taken if working with nominal variables with many levels as this could potentially
        result in many being stored in this attribute.

    unseen_levels_to_rare : bool, default = True
        If True, unseen levels in new data will be passed to rare, if set to false they will be left unchanged.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    cut_off_percent : float
        Cut off percentage (either in terms of number of rows or sum of weight) for a given
        nominal level to be considered rare.

    non_rare_levels : dict
        Created in fit. A dict of non-rare levels (i.e. levels with more than cut_off_percent weight or rows)
        that is used to identify rare levels in transform.

    rare_level_name : any
        Must be of the same type as columns.
        Label for the new nominal level that will be added to group together rare levels (as
        defined by cut_off_percent).

    record_rare_levels : bool
        Should the 'rare' levels that will be grouped together be recorded? If not they will be lost
        after the fit and the only information remaining will be the 'non'rare' levels.

    rare_levels_record_ : dict
        Only created (in fit) if record_rare_levels is True. This is dict containing a list of
        levels that were grouped into 'rare' for each column the transformer was applied to.

    weights_column : str
        Name of weights columns to use if cut_off_percent should be in terms of sum of weight
        not number of rows.

    unseen_levels_to_rare : bool
        If True, unseen levels in new data will be passed to rare, if set to false they will be left unchanged.

    training_data_levels : dict[set]
        Dictionary containing the set of values present in the training data for each column in self.columns. It
        will only exist in if unseen_levels_to_rare is set to False.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    FITS = True

    def __init__(
        self,
        columns: str | list[str] | None = None,
        cut_off_percent: float = 0.01,
        weights_column: str | None = None,
        rare_level_name: str | list[str] | None = "rare",
        record_rare_levels: bool = True,
        unseen_levels_to_rare: bool = True,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns=columns, **kwargs)

        if not isinstance(cut_off_percent, float):
            msg = f"{self.classname()}: cut_off_percent must be a float"
            raise ValueError(msg)

        if not ((cut_off_percent > 0) & (cut_off_percent < 1)):
            msg = f"{self.classname()}: cut_off_percent must be > 0 and < 1"
            raise ValueError(msg)

        self.cut_off_percent = cut_off_percent

        WeightColumnMixin.check_and_set_weight(self, weights_column)

        if not isinstance(rare_level_name, str):
            msg = f"{self.classname()}: rare_level_name must be a str"
            raise ValueError(msg)

        self.rare_level_name = rare_level_name

        if not isinstance(record_rare_levels, bool):
            msg = f"{self.classname()}: record_rare_levels must be a bool"
            raise ValueError(msg)

        self.record_rare_levels = record_rare_levels

        if not isinstance(unseen_levels_to_rare, bool):
            msg = f"{self.classname()}: unseen_levels_to_rare must be a bool"
            raise ValueError(msg)

        self.unseen_levels_to_rare = unseen_levels_to_rare

    @nw.narwhalify
    def _check_str_like_columns(self, X: FrameT) -> None:
        """check that transformer being called on only str-like columns

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to transform

        """

        str_like_columns = list(
            set(self.columns).intersection(
                set(
                    X.select(
                        ncs.string(),
                        ncs.categorical(),
                        ncs.by_dtype(nw.Object),
                    ).columns,
                ),
            ),
        )

        non_str_like_columns = set(self.columns).difference(
            set(
                str_like_columns,
            ),
        )

        if len(non_str_like_columns) != 0:
            msg = f"{self.classname()}: transformer must run on str-like columns, but got non str-like {non_str_like_columns}"
            raise TypeError(msg)

    @nw.narwhalify
    def _check_for_nulls(self, X: FrameT) -> None:
        """check that transformer being called on only non-null columns.

        Note, found including nulls to be quite complicated due to:
        - categorical variables make use of NaN not None
        - pl/nw categorical variables do not allow categories to be edited,
        so adjusting requires converting to str as interim step
        - NaNs are converted to nan, introducing complications

        As this transformer is generally used post imputation, elected to remove null
        functionality.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to transform

        """

        for c in self.columns:
            nulls_count = 0

            # pick out str_like dtypes that allow nans
            nulls_count += X.select(nw.col(c).is_null().sum()).item()

            if nulls_count != 0:
                msg = f"{self.classname()}: transformer can only fit/apply on columns without nulls, column {c} needs to be imputed first"
                raise ValueError(msg)

    @nw.narwhalify
    def fit(self, X: FrameT, y: nw.Series | None = None) -> FrameT:
        """Records non-rare levels for categorical variables.

        When transform is called, only levels records in non_rare_levels during fit will remain
        unchanged - all other levels will be grouped. If record_rare_levels is True then the
        rare levels will also be recorded.

        The label for the rare levels must be of the same type as the columns.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to identify non-rare levels from.

        y : None or or nw.Series, default = None
            Optional argument only required for the transformer to work with sklearn pipelines.

        """
        super().fit(X, y)

        if self.weights_column is not None:
            WeightColumnMixin.check_weights_column(self, X, self.weights_column)

        self._check_str_like_columns(
            X.with_columns(nw.col(col) for col in self.columns),
        )

        self._check_for_nulls(X)

        self.non_rare_levels = {}

        if self.record_rare_levels:
            self.rare_levels_record_ = {}

        native_backend = nw.get_native_namespace(X)

        weights_column = self.weights_column
        if self.weights_column is None:
            weights_column = "dummy_weights_column"
            X = X.with_columns(
                nw.new_series(
                    name=weights_column,
                    values=[1] * len(X),
                    backend=native_backend,
                ),
            )

        for c in self.columns:
            cols_w_sums = X.group_by(c, drop_null_keys=False).agg(
                nw.col(weights_column).sum(),
            )

            total_w = X.select(nw.col(weights_column).sum()).item()

            cols_w_percents = cols_w_sums.select(
                nw.col(weights_column) / total_w,
                nw.col(c),
            )

            self.non_rare_levels[c] = (
                cols_w_percents.filter(
                    nw.col(weights_column) >= self.cut_off_percent,
                )
                .get_column(c)
                .to_list()
            )

            self.non_rare_levels[c] = sorted(self.non_rare_levels[c], key=str)

            if self.record_rare_levels:
                self.rare_levels_record_[c] = (
                    cols_w_percents.filter(
                        nw.col(weights_column) < self.cut_off_percent,
                    )
                    .get_column(c)
                    .to_list()
                )

                self.rare_levels_record_[c] = sorted(
                    self.rare_levels_record_[c],
                    key=str,
                )

        if not self.unseen_levels_to_rare:
            self.training_data_levels = {}
            for c in self.columns:
                self.training_data_levels[c] = set(X.get_column(c).unique().to_list())

        return self

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Grouped rare levels together into a new 'rare' level.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to with catgeorical variables to apply rare level grouping to.

        Returns
        -------
        X : pd/pl.DataFrame
            Transformed input X with rare levels grouped for into a new rare level.

        """
        X = nw.from_native(BaseTransformer.transform(self, X))

        self._check_str_like_columns(
            X.with_columns(nw.col(col) for col in self.columns),
        )

        self._check_for_nulls(X)

        self.check_is_fitted(["non_rare_levels"])

        # copy non_rare_levels, as unseen values may be added, and transform should not
        # change the transformer state
        non_rare_levels = copy.deepcopy(self.non_rare_levels)

        if not self.unseen_levels_to_rare:
            for c in self.columns:
                unseen_vals = set(
                    X.get_column(c).unique().to_list(),
                ).difference(
                    set(self.training_data_levels[c]),
                )
                for unseen_val in unseen_vals:
                    non_rare_levels[c].append(unseen_val)

        for c in self.columns:
            categorical = False
            if str(X.schema[c]) == "Categorical":
                categorical = True
                X = X.with_columns(nw.col(c).cast(nw.String))

            non_rare_condition = nw.col(c).is_in(non_rare_levels[c])

            X = X.with_columns(
                nw.when(non_rare_condition)
                .then(
                    nw.col(c),
                )
                .otherwise(
                    nw.lit(self.rare_level_name),
                )
                .alias(
                    c,
                ),
            )

            if categorical:
                X = X.with_columns(nw.col(c).cast(nw.Categorical))

        return X


class MeanResponseTransformer(
    BaseNominalTransformer,
    WeightColumnMixin,
    BaseMappingTransformMixin,
    DropOriginalMixin,
):
    """Transformer to apply mean response encoding. This converts categorical variables to
    numeric by mapping levels to the mean response for that level.

    For a continuous or binary response the categorical columns specified will have values
    replaced with the mean response for each category.

    For an n > 1 level categorical response, up to n binary responses can be created, which in
    turn can then be used to encode each categorical column specified. This will generate up
    to n * len(columns) new columns, of with names of the form {column}_{response_level}. The
    original columns will be removed from the dataframe. This functionality is controlled using
    the 'level' parameter. Note that the above only works for a n > 1 level categorical response.
    Do not use 'level' parameter for a n = 1 level numerical response. In this case, use the standard
    mean response transformer without the 'level' parameter.

    If a categorical variable contains null values these will not be transformed.

    The same weights and prior are applied to each response level in the multi-level case.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    weights_column : str or None
        Weights column to use when calculating the mean response.

    prior : int, default = 0
        Regularisation parameter, can be thought of roughly as the size a category should be in order for
        its statistics to be considered reliable (hence default value of 0 means no regularisation).

    level : str, list or None, default = None
        Parameter to control encoding against a multi-level categorical response. For a continuous or
        binary response, leave this as None. In the multi-level case, set to 'all' to encode against every
        response level or provide a list of response levels to encode against.

    unseen_level_handling : str("Mean", "Median", "Lowest" or "Highest) or int/float, default = None
        Parameter to control the logic for handling unseen levels of the categorical features to encode in
        data when using transform method. Default value of None will output error when attempting to use transform
        on data with unseen levels in categorical columns to encode. Set this parameter to one of the options above
        in order to encode unseen levels in each categorical column with the mean, median etc. of
        each column. One can also pass an arbitrary int/float value to use for encoding unseen levels.

    return_type: Literal['float32', 'float64']
        What type to cast return column as, consider exploring float32 to save memory. Defaults to float32.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    columns : str or list
        Categorical columns to encode in the input data.

    weights_column : str or None
        Weights column to use when calculating the mean response.

    prior : int, default = 0
        Regularisation parameter, can be thought of roughly as the size a category should be in order for
        its statistics to be considered reliable (hence default value of 0 means no regularisation).

    level : str, int, float, list or None, default = None
        Parameter to control encoding against a multi-level categorical response. If None the response will be
        treated as binary or continous, if 'all' all response levels will be encoded against and if it is a list of
        levels then only the levels specified will be encoded against.

    response_levels : list
        Only created in the mutli-level case. Generated from level, list of all the response levels to encode against.

    mappings : dict
        Created in fit. A nested Dict of {column names : column specific mapping dictionary} pairs.  Column
        specific mapping dictionaries contain {initial value : mapped value} pairs.

    mapped_columns : list
        Only created in the multi-level case. A list of the new columns produced by encoded the columns in self.columns
        against multiple response levels, of the form {column}_{level}.

    transformer_dict : dict
        Only created in the mutli-level case. A dictionary of the form level : transformer containing the mean response
        transformers for each level to be encoded against.

    unseen_levels_encoding_dict: dict
        Dict containing the values (based on chosen unseen_level_handling) derived from the encoded columns to use when handling unseen levels in data passed to transform method.

    return_type: Literal['float32', 'float64']
        What type to cast return column as. Defaults to float32.

    cast_method: Literal[np.float32, np,float64]
        Store the casting method associated to return_type

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    FITS = True

    @beartype
    def __init__(
        self,
        columns: Optional[Union[str, list[str]]] = None,
        weights_column: Optional[str] = None,
        prior: int = 0,
        level: Optional[Union[float, int, str, list]] = None,
        unseen_level_handling: Optional[
            Union[float, int, Literal["mean", "median", "min", "max"]]
        ] = None,
        return_type: Literal["Float32", "Float64"] = "Float32",
        drop_original: bool = True,
        **kwargs: bool,
    ) -> None:
        if not prior >= 0:
            msg = f"{self.classname()}: prior should be positive int"
            raise ValueError(msg)

        WeightColumnMixin.check_and_set_weight(self, weights_column)

        self.prior = prior
        self.unseen_level_handling = unseen_level_handling
        self.return_type = return_type
        DropOriginalMixin.set_drop_original_column(self, drop_original=drop_original)

        self.MULTI_LEVEL = False

        if level == "all" or (isinstance(level, list)):
            self.MULTI_LEVEL = True

        # if working with single level, put into list for easier handling
        elif isinstance(level, (str, int, float)):
            level = [level]
            self.MULTI_LEVEL = True

        self.level = level

        # self.cast_method is used to cast mapping values, so uses numpy types
        if return_type == "Float64":
            self.cast_method = np.float64
        else:
            self.cast_method = np.float32

        BaseNominalTransformer.__init__(self, columns=columns, **kwargs)

    @nw.narwhalify
    def _prior_regularisation(
        self,
        group_means_and_weights: FrameT,
        column: str,
        weights_column: str,
        response_column: str,
    ) -> dict[str | float, float]:
        """Regularise encoding values by pushing encodings of infrequent categories towards the global mean.  If prior is zero this will return target_means unaltered.

        Parameters
        ----------
        group_means_and_weights: FrameT
            dataframe containing info on group means and weights for a given column

        column: str
            column to regularise

        weights_column: str
            name of weights column

        response_column: str
            name of response column

        Returns
        -------
        regularised : nw.Series
            Series of regularised encoding values
        """
        self.check_is_fitted(["global_mean"])

        prior_col = "prior_encodings"

        prior_df = group_means_and_weights.select(
            nw.col(column),
            (
                ((nw.col(response_column)) + self.global_mean * self.prior)
                / (nw.col(weights_column) + self.prior)
            ).alias(prior_col),
        )

        prior_encodings_dict = prior_df.to_dict(as_series=False)

        # return as dict
        return dict(
            zip(
                prior_encodings_dict[column],
                prior_encodings_dict[prior_col],
            ),
        )

    @nw.narwhalify
    def _fit_binary_response(
        self,
        X_y: FrameT,
        columns: list[str],
        weights_column: str,
        response_column: str,
    ) -> None:
        """Function to learn the MRE mappings for a given binary or continuous response.

        Parameters
        ----------
        X_y : pd/pl.DataFrame
            Data with catgeorical variable columns to transform, and target to encode against.

        columns : list(str)
            Post transform names of columns to be encoded. In the binary or continous case
            this is just self.columns. In the multi-level case this should be of the form
            {column_in_original_data}_{response_level}, where response_level is the level
            being encoded against in this call of _fit_binary_response.

        weights_column: str
            name of weights column

        response_column: str
            name of response column
        """
        # reuse mean imputer logic to calculate global mean
        mean_imputer = MeanImputer(
            columns=response_column,
            weights_column=weights_column,
        )
        mean_imputer.fit(X_y)
        self.global_mean = mean_imputer.impute_values_[response_column]

        X_y = X_y.with_columns(
            (nw.col(response_column) * nw.col(weights_column)).alias(response_column),
        )

        for c in columns:
            groupby_sum = X_y.group_by(c).agg(
                nw.col(response_column).sum(),
                nw.col(weights_column).sum(),
            )

            group_means_and_weights = groupby_sum.select(
                nw.col(response_column),
                nw.col(weights_column),
                nw.col(c),
            )

            self.mappings[c] = self._prior_regularisation(
                group_means_and_weights,
                column=c,
                weights_column=weights_column,
                response_column=response_column,
            )

            # to_dict changes types
            for key in self.mappings[c]:
                self.mappings[c][key] = self.cast_method(self.mappings[c][key])

    @nw.narwhalify
    def fit(self, X: FrameT, y: nw.Series) -> FrameT:
        """Identify mapping of categorical levels to mean response values.

        If the user specified the weights_column arg in when initialising the transformer
        the weighted mean response will be calculated using that column.

        In the multi-level case this method learns which response levels are present and
        are to be encoded against.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to with catgeorical variable columns to transform and also containing response_column
            column.

        y : pd/pl.Series
            Response variable or target.

        """
        BaseNominalTransformer.fit(self, X, y)

        self.mappings = {}
        self.unseen_levels_encoding_dict = {}

        if self.weights_column is not None:
            weights_column = self.weights_column
            WeightColumnMixin.check_weights_column(self, X, self.weights_column)

        else:
            native_backend = nw.get_native_namespace(X)

            weights_column = "dummy_weights_column"
            X = X.with_columns(
                nw.new_series(
                    name=weights_column,
                    values=[1] * len(X),
                    backend=native_backend.__name__,
                ),
            )

        response_null_count = y.is_null().sum()

        if response_null_count > 0:
            msg = f"{self.classname()}: y has {response_null_count} null values"
            raise ValueError(msg)

        X_y = nw.from_native(self._combine_X_y(X, y))
        response_column = "_temporary_response"

        self.response_levels = self.level

        if self.level == "all":
            self.response_levels = y.unique()

        elif self.level is not None and any(
            level not in list(y.unique()) for level in self.level
        ):
            msg = "Levels contains a level to encode against that is not present in the response."
            raise ValueError(msg)

        self.column_to_encoded_columns = {column: [] for column in self.columns}
        self.encoded_columns = []

        levels_to_iterate_through = (
            self.response_levels
            if self.response_levels is not None
            # if no levels, just create arbitrary len 1 iterable
            else ["NO_LEVELS"]
        )

        for level in levels_to_iterate_through:
            # if multi level, set up placeholder encoded columns
            # if not multi level, will just overwrite existing column
            mapping_columns_for_this_level = {
                column: column + "_" + str(level) if self.MULTI_LEVEL else column
                for column in self.columns
            }

            X_y = X_y.with_columns(
                nw.col(column).alias(mapping_columns_for_this_level[column])
                for column in mapping_columns_for_this_level
            )

            # create temporary single level response columns to individually encode
            # if nans are present then will error in previous handling, so can assume not here
            X_y = X_y.with_columns(
                (y == level if self.MULTI_LEVEL else y).alias(response_column),
            )

            self._fit_binary_response(
                X_y,
                list(mapping_columns_for_this_level.values()),
                weights_column=weights_column,
                response_column=response_column,
            )

            self.column_to_encoded_columns = {
                column: [
                    *self.column_to_encoded_columns[column],
                    mapping_columns_for_this_level[column],
                ]
                for column in self.columns
            }

        self.encoded_columns = [
            value
            for column in self.columns
            for value in self.column_to_encoded_columns[column]
        ]

        # set this attr up for BaseMappingTransformerMixin
        # this is used to cast the narwhals mapping df, so uses narwhals types
        self.return_dtypes = {col: self.return_type for col in self.encoded_columns}

        self._fit_unseen_level_handling_dict(X_y, weights_column)

        return self

    @nw.narwhalify
    def _fit_unseen_level_handling_dict(
        self,
        X_y: FrameT,
        weights_column: str,
    ) -> None:
        """Learn values for unseen levels to be mapped to, potential cases depend on unseen_level_handling attr:
        - if int/float value has been provided, this will cast to the appropriate type
        and be directly used
        - if median/mean/min/max, the appropriate weighted statistic is calculated on the mapped data, and
        cast to the appropriate type

        Parameters
        ----------
        X_y : pd/pl.DataFrame
            Data to with categorical variable columns to transform and also containing response_column
            column.

        weights_column : str
            name of weights column

        """

        if isinstance(self.unseen_level_handling, (int, float)):
            for c in self.encoded_columns:
                self.unseen_levels_encoding_dict[c] = self.cast_method(
                    self.unseen_level_handling,
                )

        elif isinstance(self.unseen_level_handling, str):
            X_temp = nw.from_native(BaseMappingTransformMixin.transform(self, X_y))

            for c in self.encoded_columns:
                if self.unseen_level_handling in ["mean", "median"]:
                    group_weights = X_temp.group_by(c).agg(
                        nw.col(weights_column).sum(),
                    )

                    if self.unseen_level_handling == "mean":
                        # reuse MeanImputer logic to calculate means
                        mean_imputer = MeanImputer(
                            columns=c,
                            weights_column=weights_column,
                        )

                        mean_imputer.fit(group_weights)

                        self.unseen_levels_encoding_dict[
                            c
                        ] = mean_imputer.impute_values_[c]

                    # else, median
                    else:
                        # reuse MedianImputer logic to calculate medians
                        median_imputer = MedianImputer(
                            columns=c,
                            weights_column=weights_column,
                        )

                        median_imputer.fit(group_weights)

                        self.unseen_levels_encoding_dict[
                            c
                        ] = median_imputer.impute_values_[c]

                # else, min or max, which don't care about weights
                else:
                    self.unseen_levels_encoding_dict[c] = getattr(
                        X_temp[c],
                        self.unseen_level_handling,
                    )()

                self.unseen_levels_encoding_dict[c] = self.cast_method(
                    self.unseen_levels_encoding_dict[c],
                )

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Transform method to apply mean response encoding stored in the mappings attribute to
        each column in the columns attribute.

        This method calls the check_mappable_rows method from BaseNominalTransformer to check that
        all rows can be mapped then transform from BaseMappingTransformMixin to apply the
        standard pd.Series.map method.

        N.B. In the mutli-level case, this method briefly overwrites the self.columns attribute, but sets
        it back to the original value at the end.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd/pl.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """

        self.check_is_fitted(["mappings", "return_dtypes"])

        # BaseTransformer.transform as we do not want to run check_mappable_rows in BaseNominalTransformer
        # (it causes complications with unseen levels and new cols, so run later)
        X = nw.from_native(BaseTransformer.transform(self, X))

        # create new columns, replacing unseen values with None
        # (they will be filled later)
        X = X.with_columns(
            nw.when(
                nw.col(old_column).is_in(self.mappings[new_column].keys()),
            )
            .then(
                nw.col(old_column),
            )
            .otherwise(
                None,
            )
            .alias(new_column)
            for old_column in self.columns
            for new_column in self.column_to_encoded_columns[old_column]
        )

        # with columns created, can now run parent transforms
        if self.unseen_level_handling:
            # do not want to run check_mappable_rows in this case, as will not like unseen values
            self.check_is_fitted(["unseen_levels_encoding_dict"])
        else:
            # temp overwrite columns attr so that check_mappable_rows runs safely
            original_columns = self.columns
            self.columns = self.encoded_columns
            X = nw.from_native(super().transform(X))
            self.columns = original_columns

        # next apply mappings
        X = nw.from_native(BaseMappingTransformMixin.transform(self, X))

        if self.unseen_level_handling:
            for old_col in self.columns:
                X = X.with_columns(
                    nw.when(
                        nw.col(new_col).is_in(list(self.mappings[new_col].values())),
                    )
                    .then(
                        nw.col(new_col),
                    )
                    .otherwise(self.unseen_levels_encoding_dict[new_col])
                    for new_col in self.column_to_encoded_columns[old_col]
                )

        columns_to_drop = [
            col for col in self.columns if col not in self.encoded_columns
        ]

        return DropOriginalMixin.drop_original_column(
            self,
            X,
            self.drop_original,
            columns_to_drop,
        )


class OrdinalEncoderTransformer(
    BaseNominalTransformer,
    BaseMappingTransformMixin,
    WeightColumnMixin,
):
    """Transformer to encode categorical variables into ascending rank-ordered integer values variables by mapping
    it's levels to the target-mean response for that level.
    Values will be sorted in ascending order only i.e. categorical level with lowest target mean response to
    be encoded as 1, the next highest value as 2 and so on.

    If a categorical variable contains null values these will not be transformed.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to transform, if the default of None is supplied all object and category
        columns in X are used.

    weights_column : str or None
        Weights column to use when calculating the mean response.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    weights_column : str or None
        Weights column to use when calculating the mean response.

    mappings : dict
        Created in fit. Dict of key (column names) value (mapping of categorical levels to numeric,
        ordinal encoded response values) pairs.

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = False

    FITS = True

    def __init__(
        self,
        columns: str | list[str] | None = None,
        weights_column: str | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        WeightColumnMixin.check_and_set_weight(self, weights_column)

        BaseNominalTransformer.__init__(self, columns=columns, **kwargs)

        # this transformer shouldn't really be used with huge numbers of levels
        # so setup to use int8 type
        # if there are more levels than this, will get a type error
        self.return_dtypes = {c: "Int8" for c in self.columns}

    def fit(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        """Identify mapping of categorical levels to rank-ordered integer values by target-mean in ascending order.

        If the user specified the weights_column arg in when initialising the transformer
        the weighted mean response will be calculated using that column.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform and response_column column
            specified when object was initialised.

        y : pd.Series
            Response column or target.

        """
        BaseNominalTransformer.fit(self, X, y)

        self.mappings = {}

        if self.weights_column is not None:
            WeightColumnMixin.check_weights_column(self, X, self.weights_column)

        response_null_count = y.isna().sum()

        if response_null_count > 0:
            msg = f"{self.classname()}: y has {response_null_count} null values"
            raise ValueError(msg)

        X_y = self._combine_X_y(X, y)
        response_column = "_temporary_response"

        for c in self.columns:
            if self.weights_column is None:
                # get the indexes of the sorted target mean-encoded dict
                _idx_target_mean = list(
                    X_y.groupby([c])[response_column]
                    .mean()
                    .sort_values(ascending=True, kind="mergesort")
                    .index,
                )

                # create a dictionary whose keys are the levels of the categorical variable
                # sorted ascending by their target-mean value
                # and whose values are ascending ordinal integers
                ordinal_encoded_dict = {
                    k: _idx_target_mean.index(k) + 1 for k in _idx_target_mean
                }

                self.mappings[c] = ordinal_encoded_dict

            else:
                groupby_sum = X_y.groupby([c])[
                    [response_column, self.weights_column]
                ].sum()

                # get the indexes of the sorted target mean-encoded dict
                _idx_target_mean = list(
                    (groupby_sum[response_column] / groupby_sum[self.weights_column])
                    .sort_values(ascending=True, kind="mergesort")
                    .index,
                )

                # create a dictionary whose keys are the levels of the categorical variable
                # sorted ascending by their target-mean value
                # and whose values are ascending ordinal integers
                ordinal_encoded_dict = {
                    k: _idx_target_mean.index(k) + 1 for k in _idx_target_mean
                }

                self.mappings[c] = ordinal_encoded_dict

        for col in self.columns:
            # if more levels than int8 type can handle, then error
            if len(self.mappings[col]) > 127:
                msg = f"{self.classname()}: column {c} has too many levels to encode"
                raise ValueError(
                    msg,
                )

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transform method to apply ordinal encoding stored in the mappings attribute to
        each column in the columns attribute. This maps categorical levels to rank-ordered integer values by target-mean in ascending order.

        This method calls the check_mappable_rows method from BaseNominalTransformer to check that
        all rows can be mapped then transform from BaseMappingTransformMixin to apply the
        standard pd.Series.map method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to with catgeorical variable columns to transform.

        Returns
        -------
        X : pd.DataFrame
            Transformed data with levels mapped to ordinal encoded values for categorical variables.

        """
        X = super().transform(X)

        return BaseMappingTransformMixin.transform(self, X)


class OneHotEncodingTransformer(
    DropOriginalMixin,
    SeparatorColumnMixin,
    BaseTransformer,
):
    """Transformer to convert categorical variables into dummy columns.

    Parameters
    ----------
    columns : str or list of strings or None, default = None
        Names of columns to transform. If the default of None is supplied all object and category
        columns in X are used.

    wanted_values: dict[str, list[str] or None , default = None
        Optional parameter to select specific column levels to be transformed. If it is None, all levels in the categorical column will be encoded. It will take the format {col1: [level_1, level_2, ...]}.

    separator : str
        Used to create dummy column names, the name will take
        the format [categorical feature][separator][category level]

    drop_original : bool, default = False
        Should original columns be dropped after creating dummy fields?

    copy : bool, default = True
        Should X be copied prior to transform? Should X be copied prior to transform? Copy argument no longer used and will be deprecated in a future release

    verbose : bool, default = True
        Should warnings/checkmarks get displayed?

    **kwargs
        Arbitrary keyword arguments passed onto sklearn OneHotEncoder.init method.

    Attributes
    ----------
    separator : str
        Separator used in naming for dummy columns.

    drop_original : bool
        Should original columns be dropped after creating dummy fields?

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    FITS = True

    def __init__(
        self,
        columns: str | list[str] | None = None,
        wanted_values: dict[str, list[str]] | None = None,
        separator: str = "_",
        drop_original: bool = False,
        copy: bool | None = None,
        verbose: bool = False,
        **kwargs: dict[str, bool],
    ) -> None:
        BaseTransformer.__init__(
            self,
            columns=columns,
            verbose=verbose,
            copy=copy,
            **kwargs,
        )

        if wanted_values is not None:
            if not isinstance(wanted_values, dict):
                msg = f"{self.classname()}: wanted_values should be a dictionary"
                raise TypeError(msg)

            for key, val_list in wanted_values.items():
                # check key is a string
                if not isinstance(key, str):
                    msg = f"{self.classname()}:  Key in 'wanted_values' should be a string"
                    raise TypeError(msg)

                # check value is a list
                if not isinstance(val_list, list):
                    msg = f"{self.classname()}: Values in the 'wanted_values' dictionary should be a list"
                    raise TypeError(msg)

                # check if each value within the list is a string
                for val in val_list:
                    if not isinstance(val, str):
                        msg = f"{self.classname()}: Entries in 'wanted_values' list should be a string"
                        raise TypeError(msg)

        self.wanted_values = wanted_values
        self.set_drop_original_column(drop_original)
        self.check_and_set_separator_column(separator)

    @nw.narwhalify
    def fit(self, X: FrameT, y: nw.Series | None = None) -> FrameT:
        """Gets list of levels for each column to be transformed. This defines which dummy columns
        will be created in transform.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to identify levels from.

        y : None
            Ignored. This parameter exists only for compatibility with sklearn.pipeline.Pipeline.

        """
        BaseTransformer.fit(self, X=X, y=y)

        # Check for nulls
        for c in self.columns:
            if X.select(nw.col(c).is_null().sum()).item() > 0:
                raise ValueError(
                    f"{self.classname()}: column %s has nulls - replace before proceeding"
                    % c,
                )

        self.categories_ = {}
        self.new_feature_names_ = {}
        # Check each field has less than 100 categories/levels
        missing_levels = {}
        for c in self.columns:
            levels = X.select(nw.col(c).unique())

            level_count = levels.select(nw.col(c).count()).item()

            if level_count > 100:
                raise ValueError(
                    f"{self.classname()}: column %s has over 100 unique values - consider another type of encoding"
                    % c,
                )

            # Set the categories_ attribute to ensure check_is_fitted works
            levels_list = levels.get_column(c).to_list()

            # for consistency
            levels_list.sort()

            # categories if 'values' is provided
            selected_values = (
                self.wanted_values.get(c, None) if self.wanted_values else None
            )

            if selected_values is None:
                final_categories = levels_list
            else:
                final_categories = selected_values

            self.categories_[c] = final_categories
            self.new_feature_names_[c] = self._get_feature_names(column=c)

            present_levels = set(X.get_column(c).unique().to_list())
            missing_levels = self._warn_missing_levels(
                present_levels,
                c,
                missing_levels,
            )

        return self

    def _warn_missing_levels(
        self,
        present_levels: list,
        c: str,
        missing_levels: dict[str, list[str]],
    ) -> dict[str, list[str]]:
        """Logs a warning for user-specifed levels that are not found in the dataset and updates "missing_levels[c]" with those missing levels.

        Parameters
        ----------
        present_levels: list
            List of levels observed in the data.
        c: str
            The column name being checked for missing user-specified levels.
        missing_levels: dict[str, list[str]]
            Dictionary containing missing user-specified levels for each column.
        Returns
        -------
        missing_levels : dict[str, list[str]]
            Dictionary updated to reflect new missing levels for column c

        """
        # print warning for missing levels
        missing_levels[c] = list(
            set(self.categories_[c]).difference(present_levels),
        )
        if len(missing_levels[c]) > 0:
            warning_msg = f"{self.classname()}: column {c} includes user-specified values {missing_levels[c]} not found in the dataset"
            warnings.warn(warning_msg, UserWarning, stacklevel=2)

        return missing_levels

    def _get_feature_names(
        self,
        column: str,
    ) -> list[str]:
        """Function to get list of features that will be output by transformer

        Parameters
        ----------
        column: str
            column to get dummy feature names for

        """

        return [
            column + self.separator + str(level) for level in self.categories_[column]
        ]

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Create new dummy columns from categorical fields.

        Parameters
        ----------
        X : pd/pl.DataFrame
            Data to apply one hot encoding to.

        Returns
        -------
        X_transformed : pd/pl.DataFrame
            Transformed input X with dummy columns derived from categorical columns added. If drop_original
            = True then the original categorical columns that the dummies are created from will not be in
            the output X.

        """

        # Check that transformer has been fit before calling transform
        self.check_is_fitted(["categories_"])

        X = nw.from_native(BaseTransformer.transform(self, X))

        missing_levels = {}
        for c in self.columns:
            # Check for nulls
            if X.select(nw.col(c).is_null().sum()).item() > 0:
                raise ValueError(
                    f"{self.classname()}: column %s has nulls - replace before proceeding"
                    % c,
                )

            # print warning for unseen levels
            present_levels = set(X.get_column(c).unique().to_list())
            unseen_levels = present_levels.difference(set(self.categories_[c]))
            if len(unseen_levels) > 0:
                warning_msg = f"{self.classname()}: column {c} has unseen categories: {unseen_levels}"
                warnings.warn(warning_msg, UserWarning, stacklevel=2)

            # print warning for missing levels
            self._warn_missing_levels(present_levels, c, missing_levels)

            dummies = X.get_column(c).to_dummies(separator=self.separator)

            # insert 0 cols for missing levels
            dummies = dummies.with_columns(
                nw.lit(0).alias(c + self.separator + str(missing_level))
                for missing_level in missing_levels[c]
            )

            wanted_dummies = self.new_feature_names_[c]

            # cast dummy columns to bool
            dummies = dummies.with_columns(
                nw.col(new_column).cast(nw.Boolean) for new_column in wanted_dummies
            )
            X = nw.concat([X, dummies.select(wanted_dummies)], how="horizontal")

        # Drop original columns if self.drop_original is True
        return DropOriginalMixin.drop_original_column(
            self,
            X,
            self.drop_original,
            self.columns,
        )



import polars as pl
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    GenericFitTests,
    GenericInitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tubular.mapping import BaseMappingTransformer


# The first part of this file builds out the tests for BaseMappingTransformer so that they can be
# imported into other test files (by not starting the class name with Test)
# The second part actually calls these tests (along with all other require tests) for the BaseMappingTransformer
class BaseMappingTransformerInitTests(GenericInitTests):
    """
    Tests for BaseMappingTransformer.init().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_no_keys_dict_error(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
    ):
        """Test that an exception is raised if mappings is a dict but with no keys."""

        kwargs = minimal_attribute_dict[self.transformer_name]
        kwargs["mappings"] = {}

        with pytest.raises(
            ValueError,
            match=f"{self.transformer_name}: mappings has no values",
        ):
            uninitialized_transformers[self.transformer_name](**kwargs)

    def test_inferred_return_dtypes(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
    ):
        "test that return_dtypes are inferred correctly if not provided"

        kwargs = minimal_attribute_dict[self.transformer_name]
        kwargs["mappings"] = {
            "a": {"a": 1, "b": 2},
            "b": {"c": True, "d": False},
            "c": {"d": 1.0, "e": 2.0},
        }
        kwargs["return_dtypes"] = None

        transformer = uninitialized_transformers[self.transformer_name](
            **kwargs,
        )

        expected = {
            "a": "Int64",
            "b": "Boolean",
            "c": "Float64",
        }

        actual = transformer.return_dtypes

        assert (
            actual == expected
        ), f"return_dtypes attr not inferred as expected, expected {expected} but got {actual}"


class BaseMappingTransformerTransformTests(GenericTransformTests):
    """
    Tests for the transform method on MappingTransformer.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_mappings_unchanged(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
        library,
    ):
        """Test that mappings is unchanged in transform."""
        df = d.create_df_3(library=library)

        mapping = {
            "b": {1: 2, 2: 3, 3: 4, 4: 5, 5: 6, 6: 7},
        }
        return_dtypes = {"b": "Float32"}

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["mappings"] = mapping
        args["return_dtypes"] = return_dtypes

        transformer = uninitialized_transformers[self.transformer_name](**args)

        # if transformer is not yet polars compatible, skip this test
        if not transformer.polars_compatible and isinstance(df, pl.DataFrame):
            return

        transformer.transform(df)

        assert (
            mapping == transformer.mappings
        ), f"{self.transformer_name}.transform has changed self.mappings unexpectedly, expected {mapping} but got {transformer.mappings}"


class TestInit(BaseMappingTransformerInitTests):
    """Tests for BaseMappingTransformer.init()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseMappingTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseMappingTransformer"


class TestTransform(BaseMappingTransformerTransformTests):
    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseMappingTransformer"

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_1(), d.create_df_1()),
    )
    def test_X_returned(self, df, expected, uninitialized_transformers):
        """Test that X is returned from transform."""
        mapping = {
            "a": {1: "a", 2: "b", 3: "c", 4: "d", 5: "e", 6: "f"},
            "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
        }

        x = BaseMappingTransformer(mappings=mapping)

        df_transformed = x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="Check X returned from transform",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseMappingTransformer"



import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    DropOriginalInitMixinTests,
    GenericFitTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
    TwoColumnListInitTests,
)
from tubular.comparison import EqualityChecker


class TestInit(
    DropOriginalInitMixinTests,
    NewColumnNameInitMixintests,
    TwoColumnListInitTests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "EqualityChecker"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "EqualityChecker"


class TestTransform(GenericTransformTests):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "EqualityChecker"

    @pytest.mark.parametrize(
        "test_dataframe",
        [d.create_df_5(), d.create_df_2(), d.create_df_9()],
    )
    def test_expected_output(self, test_dataframe):
        """Tests that the output given by EqualityChecker tranformer is as you would expect
        when all cases are neither all True nor False.
        """
        expected = test_dataframe
        expected["bool_logic"] = expected["b"] == expected["c"]

        example_transformer = EqualityChecker(
            columns=["b", "c"],
            new_column_name="bool_logic",
        )
        actual = example_transformer.transform(test_dataframe)

        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="EqualityChecker transformer does not produce the expected output",
            print_actual_and_expected=True,
        )

    @pytest.mark.parametrize(
        "test_dataframe",
        [d.create_df_5(), d.create_df_2(), d.create_df_9()],
    )
    def test_expected_output_dropped(self, test_dataframe):
        """Tests that the output given by EqualityChecker tranformer is as you would expect
        when all cases are neither all True nor False.
        """
        expected = test_dataframe.copy()
        expected["bool_logic"] = expected["b"] == expected["c"]
        expected = expected.drop(["b", "c"], axis=1)

        example_transformer = EqualityChecker(
            columns=["b", "c"],
            new_column_name="bool_logic",
            drop_original=True,
        )
        actual = example_transformer.transform(test_dataframe)

        ta.equality.assert_frame_equal_msg(
            actual=actual,
            expected=expected,
            msg_tag="EqualityChecker transformer does not produce the expected output",
            print_actual_and_expected=True,
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "EqualityChecker"



import numpy as np
import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
    WeightColumnFitMixinTests,
    WeightColumnInitMixinTests,
)
from tests.utils import assert_frame_equal_dispatch
from tubular.nominal import OrdinalEncoderTransformer


# Dataframe used exclusively in this testing script
def create_OrdinalEncoderTransformer_test_df():
    """Create DataFrame to use OrdinalEncoderTransformer tests that correct values are.

    DataFrame column a is the response, the other columns are categorical columns
    of types; object, category, int, float, bool.

    """
    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6],
            "b": ["a", "b", "c", "d", "e", "f"],
            "c": ["a", "b", "c", "d", "e", "f"],
            "d": [1, 2, 3, 4, 5, 6],
            "e": [3, 4, 5, 6, 7, 8.0],
            "f": [False, False, False, True, True, True],
        },
    )

    df["c"] = df["c"].astype("category")

    return df


class TestInit(ColumnStrListInitTests, WeightColumnInitMixinTests):
    """Tests for OrdinalEncoderTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OrdinalEncoderTransformer"


class TestFit(GenericFitTests, WeightColumnFitMixinTests):
    """Tests for OrdinalEncoderTransformer.fit()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OrdinalEncoderTransformer"

    def test_learnt_values(self):
        """Test that the ordinal encoder values learnt during fit are expected."""
        df = create_OrdinalEncoderTransformer_test_df()

        x = OrdinalEncoderTransformer(columns=["b", "d", "f"])

        x.fit(df, df["a"])

        ta.classes.test_object_attributes(
            obj=x,
            expected_attributes={
                "mappings": {
                    "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
                    "d": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6},
                    "f": {False: 1, True: 2},
                },
            },
            msg="mappings attribute",
        )

    def test_learnt_values_weight(self):
        """Test that the ordinal encoder values learnt during fit are expected if a weights column is specified."""
        df = create_OrdinalEncoderTransformer_test_df()

        x = OrdinalEncoderTransformer(weights_column="e", columns=["b", "d", "f"])

        x.fit(df, df["a"])

        ta.classes.test_object_attributes(
            obj=x,
            expected_attributes={
                "mappings": {
                    "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
                    "d": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6},
                    "f": {False: 1, True: 2},
                },
            },
            msg="mappings attribute",
        )

    def test_response_column_nulls_error(self):
        """Test that an exception is raised if nulls are present in response_column."""
        df = d.create_df_4()

        x = OrdinalEncoderTransformer(columns=["b"])

        with pytest.raises(
            ValueError,
            match="OrdinalEncoderTransformer: y has 1 null values",
        ):
            x.fit(df, df["a"])

    def test_error_for_too_many_levels(self):
        "test that transformer.transform errors for column with too many levels"
        transformer = OrdinalEncoderTransformer(columns=["a"])

        df = pd.DataFrame(
            {
                "a": list(range(1000)),
                "b": list(range(1000)),
            },
        )

        with pytest.raises(
            ValueError,
            match="OrdinalEncoderTransformer: column a has too many levels to encode",
        ):
            transformer.fit(df, df["b"])


class TestTransform(GenericTransformTests):
    """Tests for OrdinalEncoderTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OrdinalEncoderTransformer"

    def expected_df_1():
        """Expected output for ."""
        df = pd.DataFrame(
            {
                "a": [1, 2, 3, 4, 5, 6],
                "b": [1, 2, 3, 4, 5, 6],
                "c": ["a", "b", "c", "d", "e", "f"],
                "d": [1, 2, 3, 4, 5, 6],
                "e": [3.0, 4.0, 5.0, 6.0, 7.0, 8.0],
                "f": [1, 1, 1, 2, 2, 2],
            },
        )

        df["c"] = df["c"].astype("category")
        for col in ["b", "d", "f"]:
            df[col] = df[col].astype(np.int8)

        return df

    def test_learnt_values_not_modified(self):
        """Test that the mappings from fit are not changed in transform."""
        df = create_OrdinalEncoderTransformer_test_df()

        x = OrdinalEncoderTransformer(columns="b")

        x.fit(df, df["a"])

        x2 = OrdinalEncoderTransformer(columns="b")

        x2.fit(df, df["a"])

        x2.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=x.mappings,
            actual=x2.mappings,
            msg="Mean response values not changed in transform",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(
            create_OrdinalEncoderTransformer_test_df(),
            expected_df_1(),
        ),
    )
    def test_expected_output(self, df, expected):
        """Test that the output is expected from transform."""
        x = OrdinalEncoderTransformer(columns=["b", "d", "f"])

        # set the impute values dict directly rather than fitting x on df so test works with helpers
        x.mappings = {
            "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
            "d": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6},
            "f": {False: 1, True: 2},
        }

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for OrdinalEncoderTransformer outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OrdinalEncoderTransformer"



import pytest

import tests.test_data as d
from tests.base_tests import OtherBaseBehaviourTests
from tests.mapping.test_BaseCrossColumnMappingTransformer import (
    BaseCrossColumnMappingTransformerInitTests,
    BaseCrossColumnMappingTransformerTransformTests,
)


class BaseCrossColumnNumericTransformerInitTests(
    BaseCrossColumnMappingTransformerInitTests,
):
    """
    Tests for BaseCrossColumnNumericTransformer.init().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_mapping_values_not_numeric_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that an exception is raised if mappings values are not numeric."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["adjust_column"] = "c"

        args["mappings"] = {
            "b": {"a": "a", "b": "b", "c": "c", "d": "d", "e": "e", "f": "f"},
        }

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: mapping values must be numeric",
        ):
            uninitialized_transformers[self.transformer_name](**args)

    def test_inferred_return_dtypes(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
    ):
        """test that return_dtypes are inferred correctly if not provided - test is
        overloaded as these transformers can only handle numeric types
        """

        kwargs = minimal_attribute_dict[self.transformer_name]
        kwargs["mappings"] = {
            "a": {"a": 1, "b": 2},
            "c": {"d": 1.0, "e": 2.0},
        }
        kwargs["return_dtypes"] = None

        transformer = uninitialized_transformers[self.transformer_name](
            **kwargs,
        )

        expected = {
            "a": "Int64",
            "c": "Float64",
        }

        actual = transformer.return_dtypes

        assert (
            actual == expected
        ), f"return_dtypes attr not inferred as expected, expected {expected} but got {actual}"


class BaseCrossColumnNumericTransformerTransformTests(
    BaseCrossColumnMappingTransformerTransformTests,
):
    """
    Tests for the transform method on BaseCrossColumnnumericTransformer.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_adjust_col_not_numeric_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that an exception is raised if the adjust_column is not numeric."""
        df = d.create_df_2()

        mapping = {"b": {"a": 1.1, "b": 1.2, "c": 1.3, "d": 1.4, "e": 1.5, "f": 1.6}}

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["mappings"] = mapping
        args["adjust_column"] = "c"

        x = uninitialized_transformers[self.transformer_name](**args)

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: variable c must have numeric dtype.",
        ):
            x.transform(df)


class TestInit(BaseCrossColumnNumericTransformerInitTests):
    """Tests for BaseCrossColumnNumeicTransformer.init()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseCrossColumnNumericTransformer"


class TestTransform(BaseCrossColumnNumericTransformerTransformTests):
    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseCrossColumnNumericTransformer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseCrossColumnNumericTransformer"



import narwhals as nw
import numpy as np
import pytest

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
)
from tests.imputers.test_BaseImputer import (
    GenericImputerTransformTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.imputers import NearestMeanResponseImputer


class TestInit(ColumnStrListInitTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NearestMeanResponseImputer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NearestMeanResponseImputer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_null_values_in_response_error(self, library):
        """Test an error is raised if the response column contains null entries."""
        df = d.create_df_3(library=library)

        transformer = NearestMeanResponseImputer(columns=["b"])

        with pytest.raises(
            ValueError,
            match="NearestMeanResponseImputer: y has 1 null values",
        ):
            transformer.fit(df, df["a"])

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_columns_with_no_nulls_warning(self, library):
        """Test a warning is raised if a non-response column contains no nulls."""
        df = d.create_numeric_df_1(library=library)

        transformer = NearestMeanResponseImputer(columns=["c"])

        with pytest.warns(
            UserWarning,
            match="NearestMeanResponseImputer: Column c has no missing values, this transformer will have no effect for this column.",
        ):
            transformer.fit(df, df["c"])

        expected_impute_values = {"c": None}
        assert (
            transformer.impute_values_ == expected_impute_values
        ), f"impute_values_ attr not as expected, expected {expected_impute_values} but got {transformer.impute_values_}"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values(self, library):
        """Test that the nearest response values learnt during fit are expected."""
        df = d.create_numeric_df_2(library=library)

        transformer = NearestMeanResponseImputer(columns=["b", "c"])

        transformer.fit(df, df["a"])

        assert transformer.impute_values_ == {
            "b": np.float64(3),
            "c": np.float64(2),
        }, "impute_values_ attribute"


class TestTransform(
    GenericTransformTests,
    GenericImputerTransformTests,
):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NearestMeanResponseImputer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("fit_col", "transform_col"),
        [
            # try a few types, with and without nulls in transform col
            ([1, 2, 3], [1.0, np.nan, np.nan]),
            ([4, 5, 6], [7, 8, 9]),
            (["a", "b", "c"], ["a", None, "d"]),
            (["c", "d", "e"], ["f", "g", "h"]),
            ([4.0, 5.0, 6.0], [8.0, np.nan, 6.0]),
            ([1.0, 2.0, 3.0], [4.0, 3.0, 2.0]),
            ([True, False, False], [True, True, None]),
            ([True, False, True], [True, False, True]),
        ],
    )
    def test_no_effect_when_fit_on_null_free_col(self, fit_col, transform_col, library):
        "test that when transformer fits on a col with no nulls, transform has no effect"

        df_fit_dict = {
            "a": fit_col,
            "b": [1] * len(fit_col),
        }

        df_fit = dataframe_init_dispatch(df_fit_dict, library=library)

        df_transform_dict = {
            "a": transform_col,
        }

        df_transform = dataframe_init_dispatch(df_transform_dict, library=library)

        transformer = NearestMeanResponseImputer(columns=["a"])

        transformer.fit(df_fit, df_fit["b"])

        df_transform = nw.from_native(df_transform)

        expected_output = df_transform.clone().to_native()

        df_transform = nw.to_native(df_transform)

        output = transformer.transform(df_transform)

        assert_frame_equal_dispatch(output, expected_output)



from __future__ import annotations

import inspect
import pkgutil
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import tubular.base as base
from tests.test_data import (
    create_is_between_dates_df_1,
    create_numeric_df_1,
    create_numeric_df_2,
    create_object_df,
)

if TYPE_CHECKING:
    import pandas as pd

"""
How To Use This Testing Framework
-----------------------------------

Fixtures in this file are used in tests for shared behaviour which can be run on multiple transformers through inheritance.
setup_class is defined in each test class to define cls.transformer_name as the name of the transformer being tested.

Test methods for shared behaviour are defined in base_tests.py or child test classes which are themselves inherited.  The
fixtures in this file can be used in these tests to access initiated/ uninitiated transformers and minimum attributes for
intiatialization.

Test classes which can import tests for shared behaviour inherit from an appropriate parent and define cls.transformer_name
in setup_class so that these fixtures will retrieve the appropriate transformer instance. setup_class is called by pytest
at a class level before its test methods are run https://docs.pytest.org/en/8.0.x/how-to/xunit_setup.html.

New transformers will need to be added to minimal_attribute_dict.
 """


def get_all_classes(
    wanted_module: str | None = None,
) -> dict[str, base.BaseTransformer]:
    """Method to call the weighted_quantile method and prepare the outputs.

    If there are no None values in the supplied quantiles then the outputs from weighted_quantile
    are returned as is. If there are then prepare_quantiles removes the None values before
    calling weighted_quantile and adds them back into the output, in the same position, after
    calling.

    Parameters
    ----------
    wanted_module : str or None
        str indicating which module to load classes from (e.g. 'tubular.dates'). If none, loads from all modules.

    Returns
    -------
    all_classes : dict[str, BaseTransformer]
        Dictionary containing classes in format {transformer_name:transformer}

    """
    root = str(Path(__file__).parent.parent)

    all_classes = []
    modules_to_ignore = [
        "tests",
        "conftest",
        "setup",
    ]

    for _importer, modname, _ispkg in pkgutil.walk_packages(
        path=[root],
    ):
        if wanted_module and modname != wanted_module:
            continue
        mod_parts = modname.split(".")
        if any(part in modules_to_ignore for part in mod_parts) or "_" in modname:
            continue
        module = import_module(modname)
        classes = inspect.getmembers(module, inspect.isclass)

        classes = [
            (name, transformer)
            for name, transformer in classes
            if issubclass(transformer, base.BaseTransformer)
        ]

        all_classes.extend(classes)

    return set(all_classes)


@pytest.fixture()
def minimal_attribute_dict():
    """defines minmal attributes (values) needed to initiate each transformer named (key).
    New transformers need to be added here"""
    return {
        "ArbitraryImputer": {
            "columns": ["b"],
            "impute_value": 1,
        },
        "BaseCappingTransformer": {
            "capping_values": {"a": [0.1, 0.2]},
        },
        "BaseCrossColumnMappingTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            "adjust_column": "b",
        },
        "BaseCrossColumnNumericTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            "adjust_column": "b",
        },
        "BaseGenericDateTransformer": {
            "columns": ["a"],
            "new_column_name": "bla",
        },
        "BaseDatetimeTransformer": {
            "columns": ["a"],
            "new_column_name": "bla",
        },
        "BaseDateTwoColumnTransformer": {
            "columns": ["a", "b"],
            "new_column_name": "bla",
        },
        "BaseImputer": {
            "columns": ["b"],
        },
        "BaseMappingTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            # this arg is dependent on inputs, so
            # may not generalise well
            "return_dtypes": {"a": "Int8"},
        },
        "BaseMappingTransformMixin": {
            "columns": ["a"],
        },
        "BaseNominalTransformer": {
            "columns": ["b"],
        },
        "BaseNumericTransformer": {
            "columns": ["a", "b"],
        },
        "BaseTransformer": {
            "columns": ["a"],
        },
        "BetweenDatesTransformer": {
            "new_column_name": "e",
            "columns": ["a", "c", "b"],
        },
        "CappingTransformer": {
            "capping_values": {"a": [0.1, 0.2]},
        },
        "ColumnDtypeSetter": {
            "columns": ["a"],
            "dtype": str,
        },
        "CrossColumnAddTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            "adjust_column": "b",
        },
        "CrossColumnMappingTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            "adjust_column": "b",
        },
        "CrossColumnMultiplyTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            "adjust_column": "b",
        },
        "CutTransformer": {
            "new_column_name": "b",
            "column": "a",
            "cut_kwargs": {
                "bins": 3,
            },
        },
        "DataFrameMethodTransformer": {
            "columns": ["a", "b"],
            "new_column_names": "f",
            "pd_method_name": "sum",
            "drop_original": True,
        },
        "DateDifferenceTransformer": {
            "columns": ["a", "b"],
            "new_column_name": "new_column",
        },
        "DateDiffLeapYearTransformer": {
            "columns": ["a", "b"],
            "new_column_name": "new_column",
        },
        "DatetimeInfoExtractor": {
            "columns": ["a"],
        },
        "DatetimeSinusoidCalculator": {
            "columns": ["a"],
            "method": ["sin"],
            "units": "month",
        },
        "EqualityChecker": {
            "columns": ["a", "b"],
            "new_column_name": "c",
            "drop_original": True,
        },
        "GroupRareLevelsTransformer": {
            "columns": ["b"],
        },
        "InteractionTransformer": {
            "columns": ["a", "b"],
        },
        "LogTransformer": {
            "columns": ["a"],
        },
        "MappingTransformer": {
            "mappings": {"a": {1: 2, 3: 4}},
            # this arg is dependent on inputs, so
            # may not generalise well
            "return_dtypes": {"a": "Int8"},
        },
        "MeanImputer": {
            "columns": ["b"],
        },
        "MeanResponseTransformer": {
            "columns": ["b"],
        },
        "MedianImputer": {
            "columns": ["b"],
        },
        "ModeImputer": {
            "columns": ["b"],
        },
        "NearestMeanResponseImputer": {
            "columns": ["b"],
        },
        "NominalToIntegerTransformer": {
            "columns": ["b"],
        },
        "NullIndicator": {
            "columns": ["a"],
        },
        "OneDKmeansTransformer": {
            "columns": "a",
            "new_column_name": "new",
        },
        "OneHotEncodingTransformer": {
            "columns": ["a", "b"],
            "drop_original": True,
            "separator": "-",
        },
        "OrdinalEncoderTransformer": {
            "columns": ["b"],
        },
        "OutOfRangeNullTransformer": {
            "capping_values": {"a": [0.1, 0.2]},
        },
        "PCATransformer": {
            "columns": ["a", "c"],
        },
        "ScalingTransformer": {
            "scaler_type": "standard",
            "columns": ["a", "b"],
        },
        "SeriesDtMethodTransformer": {
            "new_column_name": "new_column",
            "pd_method_name": "month",
            "columns": "b",
        },
        "SeriesStrMethodTransformer": {
            "columns": ["b"],
            "new_column_name": "a",
            "pd_method_name": "split",
        },
        "SetValueTransformer": {
            "value": 1,
            "columns": ["a"],
        },
        "StringConcatenator": {
            "columns": ["a", "b"],
            "new_column_name": "c",
            "separator": "-",
        },
        "ToDatetimeTransformer": {
            "new_column_name": "b",
            "column": "a",
        },
        "TwoColumnOperatorTransformer": {
            "columns": ["a", "b"],
            "new_column_name": "c",
            "pd_method_name": "add",
        },
    }


@pytest.fixture()
def minimal_dataframe_lookup(request) -> dict[str, pd.DataFrame]:
    """links transformers to minimal dataframes needed to successfully run transformer. There is logic to do this automatically by module, so function will only need to be edited where either:
    - a new module that operates primarily on non-numeric columns is added
    - a new transformer is added to an existing module that breaks the pattern of that module, e.g. a transformer in dates.py that operates on numeric columns

    Returns
    -------

    min_df_dict: dict[str, pd.DataFrame]
        dictionary mapping transformers to minimal dataframes that they can successfully run on

    """

    # setup to default to pandas if not provided
    library = getattr(request, "param", "pandas")

    num_df = create_numeric_df_1(library=library)
    nan_df = create_numeric_df_2(library=library)
    object_df = create_object_df(library=library)
    date_df = create_is_between_dates_df_1(library=library)

    # generally most transformers will work with num_df
    min_df_dict = {x[0]: num_df for x in get_all_classes()}

    # override dict value with date_df for transformers in tubular.dates
    date_transformers = [x[0] for x in get_all_classes(wanted_module="tubular.dates")]
    for transformer in date_transformers:
        min_df_dict[transformer] = date_df

    # override dict value for transformers that run on object type columns
    object_modules = ["tubular.mapping", "tubular.nominal", "tubular.strings"]
    object_transformers = []
    for module in object_modules:
        object_transformers = object_transformers + [
            x[0] for x in get_all_classes(wanted_module=module)
        ]

    for transformer in object_transformers:
        min_df_dict[transformer] = object_df

    # Some may require further manual overwrites
    other_num_transformers = [
        "CrossColumnMultiplyTransformer",
        "CrossColumnAddTransformer",
        "BaseCrossColumnNumericTransformer",
        # prevent below two from being imported from import in nominal file
        "MedianImputer",
        "MeanImputer",
    ]
    for transformer in other_num_transformers:
        min_df_dict[transformer] = num_df

    # Some transformers require missing values to work
    other_nan_transformers = [
        "NearestMeanResponseImputer",
    ]
    for transformer in other_nan_transformers:
        min_df_dict[transformer] = nan_df

    return min_df_dict


@pytest.fixture()
def initialized_transformers(minimal_attribute_dict):
    """dictionary of {transformer name : initiated transformer} pairs for all transformers"""
    return {x[0]: x[1](**minimal_attribute_dict[x[0]]) for x in get_all_classes()}


@pytest.fixture()
def uninitialized_transformers():
    """dictionary of {transformer name : uninitiated transformer} pairs for all transformers"""
    return {x[0]: x[1] for x in get_all_classes()}



import pytest

from tests.test_data import create_df_1
from tests.utils import assert_frame_equal_dispatch
from tubular.mixins import DropOriginalMixin


class TestSetDropOriginalColumn:
    "tests for DropOriginalMixin.set_drop_original_column"

    @pytest.mark.parametrize("drop_orginal_column", (0, "a", ["a"], {"a": 10}, None))
    def test_drop_column_arg_errors(
        self,
        drop_orginal_column,
    ):
        """Test that appropriate errors are throwm for non boolean arg."""

        obj = DropOriginalMixin()

        with pytest.raises(
            TypeError,
            match="DropOriginalMixin: drop_original should be bool",
        ):
            obj.set_drop_original_column(drop_original=drop_orginal_column)


class TestDropOriginalColumn:
    "tests for DropOriginalMixin.drop_original_column"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize("drop_original", [True, False])
    def test_drop_original_arg_handling(
        self,
        library,
        drop_original,
    ):
        """Test transformer drops/keeps original columns when specified/not specified."""

        df = create_df_1(library=library)

        obj = DropOriginalMixin()

        columns = list(df.columns)

        df_transformed = obj.drop_original_column(
            df,
            drop_original=drop_original,
            columns=columns,
        )

        remaining_cols = df_transformed.columns

        if drop_original:
            for col in columns:
                assert col not in remaining_cols, "original columns not dropped"

        else:
            for col in columns:
                assert col in remaining_cols, "original columns not kept"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_other_columns_not_modified(
        self,
        library,
    ):
        """Test transformer does not modify unspecified columns."""

        df = create_df_1(library=library)

        obj = DropOriginalMixin()

        drop_original = False

        columns = ["a"]

        df_transformed = obj.drop_original_column(
            df,
            drop_original=drop_original,
            columns=columns,
        )

        other_columns = list(set(df.columns) - set(columns))

        assert_frame_equal_dispatch(df[other_columns], df_transformed[other_columns])



from sklearn.pipeline import Pipeline

from tubular.capping import CappingTransformer, OutOfRangeNullTransformer
from tubular.imputers import (
    ArbitraryImputer,
    MeanImputer,
    MedianImputer,
    ModeImputer,
    NearestMeanResponseImputer,
    NullIndicator,
)
from tubular.mapping import MappingTransformer
from tubular.nominal import (
    GroupRareLevelsTransformer,
    MeanResponseTransformer,
    NominalToIntegerTransformer,
    OneHotEncodingTransformer,
)
from tubular.numeric import LogTransformer


class TubularPipelineGenerator:
    """Class to generate pipelines containing any combination of the tubular transformers that are also found in fubular."""

    def __init__(self) -> None:
        self.all_tubular_transformers = [
            "CappingTransformer",
            "OutOfRangeNullTransformer",
            "ArbitraryImputer",
            "MedianImputer",
            "MeanImputer",
            "ModeImputer",
            "NearestMeanResponseImputer",
            "NullIndicator",
            "MappingTransformer",
            "NominalToIntegerTransformer",
            "MeanResponseTransformer",
            "GroupRareLevelsTransformer",
            "OneHotEncodingTransformer",
            "LogTransformer",
        ]

    def get_CappingTransformer(self) -> CappingTransformer:
        return CappingTransformer(capping_values={"AveOccup": [1, 5.5]})

    def get_OutOfRangeNullTransformer(self) -> OutOfRangeNullTransformer:
        return OutOfRangeNullTransformer(capping_values={"AveOccup": [1, 5.5]})

    def get_ArbitraryImputer(self) -> ArbitraryImputer:
        return ArbitraryImputer(
            columns=["HouseAge_1", "AveOccup_1", "Population_1"],
            impute_value=-1,
        )

    def get_MedianImputer(self) -> MedianImputer:
        return MedianImputer(columns=["HouseAge_2", "AveOccup_2", "Population_2"])

    def get_MeanImputer(self) -> MeanImputer:
        return MeanImputer(columns=["HouseAge_3", "AveOccup_3", "Population_3"])

    def get_ModeImputer(self) -> ModeImputer:
        return ModeImputer(columns=["HouseAge_4", "AveOccup_4", "Population_4"])

    def get_NearestMeanResponseImputer(self) -> NearestMeanResponseImputer:
        return NearestMeanResponseImputer(
            columns=["HouseAge_5", "AveOccup_5", "Population_5"],
        )

    def get_NullIndicator(self) -> NullIndicator:
        return NullIndicator(columns=["HouseAge_6", "AveOccup_6", "Population_6"])

    def get_MappingTransformer(self) -> MappingTransformer:
        return MappingTransformer(mappings={"categorical_1": {"a": "c", "b": "d"}})

    def get_NominalToIntegerTransformer(self) -> NominalToIntegerTransformer:
        return NominalToIntegerTransformer(columns=["categorical_2"])

    def get_MeanResponseTransformer(self) -> MeanResponseTransformer:
        return MeanResponseTransformer(columns=["categorical_3"])

    def get_GroupRareLevelsTransformer(self) -> GroupRareLevelsTransformer:
        return GroupRareLevelsTransformer(columns=["categorical_4"])

    def get_OneHotEncodingTransformer(self) -> OneHotEncodingTransformer:
        return OneHotEncodingTransformer(columns=["categorical_ohe"])

    def get_LogTransformer(self) -> LogTransformer:
        return LogTransformer(columns=["HouseAge_7", "AveOccup_7", "Population_7"])

    def generate_pipeline(
        self,
        transformers_to_include: list = None,
        verbose: bool = False,
    ) -> Pipeline:
        if not transformers_to_include:
            transformers_to_include = self.all_tubular_transformers

        steps = [
            (transformer, getattr(self, f"get_{transformer}")())
            for transformer in transformers_to_include
        ]

        return Pipeline(steps, verbose=verbose)



import copy
import re

import narwhals as nw
import numpy as np
import polars as pl
import pytest

from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    GenericFitTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
)
from tests.test_data import create_date_diff_different_dtypes


class GenericDatesMixinTransformTests:
    """Generic tests for Dates Transformers"""

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=["minimal_dataframe_lookup"],
    )
    @pytest.mark.parametrize(
        ("bad_value", "bad_type"),
        [
            (1, "Int64"),
            ("a", "String"),
            (np.nan, "Float64"),
        ],
    )
    def test_non_datetypes_error(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
        minimal_dataframe_lookup,
        bad_value,
        bad_type,
    ):
        "Test that transform raises an error if columns contains non date types"

        args = minimal_attribute_dict[self.transformer_name].copy()
        columns = args["columns"]

        transformer = uninitialized_transformers[self.transformer_name](
            **args,
        )

        df = copy.deepcopy(minimal_dataframe_lookup[self.transformer_name])

        # if transformer is not yet polars compatible, skip this test
        if not transformer.polars_compatible and isinstance(df, pl.DataFrame):
            return

        for i in range(len(columns)):
            col = columns[i]
            bad_df = nw.from_native(df).clone()
            bad_df = bad_df.with_columns(
                nw.lit(bad_value).cast(getattr(nw, bad_type)).alias(col),
            )

            msg = rf"{col} type should be in \['datetime64', 'date32\[pyarrow\]'\] but got {bad_type}"

            with pytest.raises(
                TypeError,
                match=msg,
            ):
                transformer.transform(nw.to_native(bad_df))

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("columns, datetime_col"),
        [
            (["date_col_1", "datetime_col_2"], 1),
            (["datetime_col_1", "date_col_2"], 0),
        ],
    )
    def test_mismatched_datetypes_error(
        self,
        columns,
        datetime_col,
        uninitialized_transformers,
        minimal_attribute_dict,
        library,
    ):
        "Test that transform raises an error if one column is a date and one is datetime"
        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = columns

        transformer = uninitialized_transformers[self.transformer_name](
            **args,
        )

        df = create_date_diff_different_dtypes(library=library)

        # if transformer is not yet polars compatible, skip this test
        if not transformer.polars_compatible and isinstance(df, pl.DataFrame):
            return

        present_types = (
            {"datetime64", r"date32[pyarrow]"}
            if datetime_col == 0
            else {r"date32[pyarrow]", "datetime64"}
        )
        msg = re.escape(
            f"Columns fed to datetime transformers should be ['datetime64', 'date32[pyarrow]'] and have consistent types, but found {present_types}. Please use ToDatetimeTransformer to standardise",
        )
        with pytest.raises(
            TypeError,
            match=msg,
        ):
            transformer.transform(df)


class TestInit(
    NewColumnNameInitMixintests,
    DropOriginalInitMixinTests,
    ColumnStrListInitTests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseGenericDateTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseGenericDateTransformer"


class TestTransform(GenericTransformTests, GenericDatesMixinTransformTests):
    """Tests for BaseGenericDateTransformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseGenericDateTransformer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseGenericDateTransformer"



import numpy as np
import polars as pl
import pytest

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.misc import SetValueTransformer


def expected_df_1(library, value):
    """Expected output of test_value_set_in_transform."""

    df_dict = {
        "a": [value] * 7,
        "b": [value] * 7,
        "c": ["a", "b", "c", "d", "e", "f", None],
    }

    df = dataframe_init_dispatch(df_dict, library)
    if library == "pandas":
        df["c"] = df["c"].astype("category")
    elif library == "polars":
        df = df.with_columns(df["c"].cast(pl.Categorical))
        # polars automatically downcasts to int32 in transformer
        if isinstance(value, int):
            df = df.with_columns(df[["a", "b"]].cast(pl.Int32))
    return df


class TestInit(ColumnStrListInitTests):
    """Generic tests for SetValueTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SetValueTransformer"


class TestFit(GenericFitTests):
    """Generic tests for SetValueTransformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SetValueTransformer"


class TestTransform(GenericTransformTests):
    """Tests for SetValueTransformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SetValueTransformer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize("value", ["a", 1, 1.0, None, np.nan])
    def test_value_set_in_transform(self, library, value):
        """Test that transform sets the value as expected."""

        df = d.create_df_2(library)

        x = SetValueTransformer(columns=["a", "b"], value=value)

        df_transformed = x.transform(df)

        expected = expected_df_1(library, value)

        assert_frame_equal_dispatch(
            df1=df_transformed,
            df2=expected,
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for SetValueTransformer behaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "SetValueTransformer"



import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.mapping.test_BaseMappingTransformer import (
    BaseMappingTransformerInitTests,
    BaseMappingTransformerTransformTests,
)


class BaseCrossColumnMappingTransformerInitTests(BaseMappingTransformerInitTests):
    """
    Tests for BaseCrossColumnMappingTransformer.init().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_adjust_columns_non_string_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that an exception is raised if adjust_column is not a string."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["adjust_column"] = 1

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: adjust_column should be a string",
        ):
            uninitialized_transformers[self.transformer_name](**args)

    def test_inferred_return_dtypes(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
    ):
        """test that return_dtypes are inferred correctly if not provided - test is
        overloaded as these transformers require OrderedDict for multicolumn mapping
        """

        kwargs = minimal_attribute_dict[self.transformer_name]
        kwargs["mappings"] = {
            "a": {"a": 1, "b": 2},
        }
        kwargs["return_dtypes"] = None

        transformer = uninitialized_transformers[self.transformer_name](
            **kwargs,
        )

        expected = {
            "a": "Int64",
        }

        actual = transformer.return_dtypes

        assert (
            actual == expected
        ), f"return_dtypes attr not inferred as expected, expected {expected} but got {actual}"


class BaseCrossColumnMappingTransformerTransformTests(
    BaseMappingTransformerTransformTests,
):
    """
    Tests for the transform method on BaseCrossColumnMappingTransformer.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def expected_df_2():
        """Expected output from test_non_specified_values_unchanged."""
        return pd.DataFrame(
            {"b": ["a", "b", "c", "d", "e", "f"]},
        )

    def test_adjust_col_not_in_x_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that an exception is raised if the adjust_column is not present in the dataframe."""
        df = d.create_df_1()

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["adjust_column"] = "c"

        args["mappings"] = {
            "b": {"a": 1.1, "b": 1.2, "c": 1.3, "d": 1.4, "e": 1.5, "f": 1.6},
        }

        x = uninitialized_transformers[self.transformer_name](**args)

        with pytest.raises(
            ValueError,
            match=f"{self.transformer_name}: variable c is not in X",
        ):
            x.transform(df)

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_1(), expected_df_2()),
    )
    def test_non_specified_values_unchanged(
        self,
        df,
        expected,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that values not specified in mappings are left unchanged in transform."""
        mapping = {"b": {"a": 1.1, "b": 1.2}}

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["mappings"] = mapping
        args["adjust_column"] = "a"

        x = uninitialized_transformers[self.transformer_name](**args)

        df_transformed = x.transform(df)

        ta.equality.assert_series_equal_msg(
            actual=df_transformed["b"],
            expected=expected["b"],
            msg_tag=f"expected output from {self.transformer_name}",
        )


class TestInit(BaseCrossColumnMappingTransformerInitTests):
    """Tests for BaseCrossColumnMappingTransformer.init()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseCrossColumnMappingTransformer"


class TestTransform(BaseCrossColumnMappingTransformerTransformTests):
    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseCrossColumnMappingTransformer"



from contextlib import suppress
from importlib.metadata import version

from tubular import (
    base,
    capping,
    dates,
    imputers,
    mapping,
    misc,
    nominal,
    numeric,
    strings,
)

with suppress(ModuleNotFoundError):
    __version__ = version("tubular")



import narwhals as nw
import pytest
from test_BaseNominalTransformer import GenericNominalTransformTests

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericFitTests,
    GenericTransformTests,
    SeparatorInitMixintests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.nominal import OneHotEncodingTransformer


class TestInit(
    SeparatorInitMixintests,
    DropOriginalInitMixinTests,
    ColumnStrListInitTests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OneHotEncodingTransformer"

    # Tests for wanted_values parameter

    @pytest.mark.parametrize(
        "values",
        ["a", ["a", "b"], 123, True],
    )
    def test_wanted_values_is_dict(self, values, minimal_attribute_dict):
        args = minimal_attribute_dict[self.transformer_name]
        args["wanted_values"] = values

        with pytest.raises(
            TypeError,
            match="OneHotEncodingTransformer: wanted_values should be a dictionary",
        ):
            OneHotEncodingTransformer(**args)

    @pytest.mark.parametrize(
        "values",
        [
            {1: ["a", "b"]},
            {True: ["a"]},
            {("a",): ["b", "c"]},
        ],
    )
    def test_wanted_values_key_is_str(self, values, minimal_attribute_dict):
        args = minimal_attribute_dict[self.transformer_name]
        args["wanted_values"] = values

        with pytest.raises(
            TypeError,
            match="OneHotEncodingTransformer:  Key in 'wanted_values' should be a string",
        ):
            OneHotEncodingTransformer(**args)

    @pytest.mark.parametrize(
        "values",
        [
            {"a": "b"},
            {"a": ("a", "b")},
            {"a": True},
            {"a": 123},
        ],
    )
    def test_wanted_values_value_is_list(self, values, minimal_attribute_dict):
        args = minimal_attribute_dict[self.transformer_name]
        args["wanted_values"] = values

        with pytest.raises(
            TypeError,
            match="OneHotEncodingTransformer: Values in the 'wanted_values' dictionary should be a list",
        ):
            OneHotEncodingTransformer(**args)

    @pytest.mark.parametrize(
        "values",
        [
            {"a": ["b", 123]},
            {"a": ["b", True]},
            {"a": ["b", None]},
            {"a": ["b", ["a", "b"]]},
        ],
    )
    def test_wanted_values_entries_are_str(self, values, minimal_attribute_dict):
        args = minimal_attribute_dict[self.transformer_name]
        args["wanted_values"] = values

        with pytest.raises(
            TypeError,
            match="OneHotEncodingTransformer: Entries in 'wanted_values' list should be a string",
        ):
            OneHotEncodingTransformer(**args)


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OneHotEncodingTransformer"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_nulls_in_X_error(self, library):
        """Test that an exception is raised if X has nulls in column to be fit on."""
        df = d.create_df_2(library=library)

        transformer = OneHotEncodingTransformer(columns=["b", "c"])

        with pytest.raises(
            ValueError,
            match="OneHotEncodingTransformer: column b has nulls - replace before proceeding",
        ):
            transformer.fit(df)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_fit_missing_levels_warning(self, library):
        """Test OneHotEncodingTransformer.fit triggers a warning for missing levels."""
        df = d.create_df_1(library=library)

        transformer = OneHotEncodingTransformer(
            columns=["b"],
            wanted_values={"b": ["f", "g"]},
        )

        with pytest.warns(
            UserWarning,
            match=(
                r"OneHotEncodingTransformer: column b includes user-specified values \['g'\] not found in the dataset"
            ),
        ):
            transformer.fit(df)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_fields_with_over_100_levels_error(self, library):
        """Test that OneHotEncodingTransformer.fit on fields with more than 100 levels raises error."""
        df_dict = {"a": [1] * 101, "b": list(range(101))}

        df = dataframe_init_dispatch(library=library, dataframe_dict=df_dict)

        transformer = OneHotEncodingTransformer(columns=["a", "b"])

        with pytest.raises(
            ValueError,
            match="OneHotEncodingTransformer: column b has over 100 unique values - consider another type of encoding",
        ):
            transformer.fit(df)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_fit_no_warning_if_all_wanted_values_present(self, library, recwarn):
        """Test that OneHotEncodingTransformer.fit does NOT raise a warning when all levels in wanted_levels are present in the data."""
        df = d.create_df_1(library=library)

        transformer = OneHotEncodingTransformer(
            columns=["b"],
            wanted_values={"b": ["a", "b", "c", "d", "e", "f"]},
        )

        transformer.fit(df)
        assert (
            len(recwarn) == 0
        ), "OneHotEncodingTransformer.fit is raising unexpected warnings"


class TestTransform(
    DropOriginalTransformMixinTests,
    GenericNominalTransformTests,
    GenericTransformTests,
):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OneHotEncodingTransformer"

    def create_OneHotEncoderTransformer_test_df_1(self, library="pandas"):
        """Create DataFrame to test OneHotEncoderTransformer

        binary columns are representative of transformed output of column b

        Parameters
        ----------
        library : str
            Whether to return polars of pandas df

        """

        df_dict = {
            "a": [4, 2, 2, 1, 3],
            "b": ["x", "z", "y", "x", "x"],
            "c": ["c", "a", "a", "c", "b"],
            "b_x": [1.0, 0.0, 0.0, 1.0, 1.0],
            "b_y": [0.0, 0.0, 1.0, 0.0, 0.0],
            "b_z": [0.0, 1.0, 0.0, 0.0, 0.0],
        }

        df = dataframe_init_dispatch(library=library, dataframe_dict=df_dict)

        df = nw.from_native(df)
        df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

        return df.to_native()

    def create_OneHotEncoderTransformer_test_df_2(self, library="pandas"):
        """Create DataFrame to test OneHotEncoderTransformer

        binary columns are representative of transformed output of all columns

        Parameters
        ----------
        library : str
            Whether to return polars of pandas df

        """

        df_dict = {
            "a": [1, 5, 2, 3, 3],
            "b": ["w", "w", "z", "y", "x"],
            "c": ["a", "a", "c", "b", "a"],
            "a_1": [1.0, 0.0, 0.0, 0.0, 0.0],
            "a_2": [0.0, 0.0, 1.0, 0.0, 0.0],
            "a_3": [0.0, 0.0, 0.0, 1.0, 1.0],
            "a_4": [0.0, 0.0, 0.0, 0.0, 0.0],
            "b_x": [0.0, 0.0, 0.0, 0.0, 1.0],
            "b_y": [0.0, 0.0, 0.0, 1.0, 0.0],
            "b_z": [0.0, 0.0, 1.0, 0.0, 0.0],
            "c_a": [1.0, 1.0, 0.0, 0.0, 1.0],
            "c_b": [0.0, 0.0, 0.0, 1.0, 0.0],
            "c_c": [0.0, 0.0, 1.0, 0.0, 0.0],
        }

        df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

        df = nw.from_native(df)
        df = df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

        return df.to_native()

    def test_non_mappable_rows_exception_raised(self):
        """Test inherited from GenericBaseNominalTransformerTests needs to be overwritten,
        inherited test tests the mapping attribute, which OHE transfomer doesn't have.
        """

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_non_numeric_column_error_1(self, library):
        """Test that transform will raise an error if a column to transform has nulls."""
        df_train = d.create_df_1(library=library)
        df_test = d.create_df_2(library=library)

        transformer = OneHotEncodingTransformer(columns=["b"])

        transformer.fit(df_train)

        with pytest.raises(
            ValueError,
            match="OneHotEncodingTransformer: column b has nulls - replace before proceeding",
        ):
            transformer.transform(df_test)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_expected_output(self, library):
        """Test that OneHotEncodingTransformer.transform encodes the feature correctly.

        Also tests that OneHotEncodingTransformer.transform does not modify unrelated columns.
        """
        # transformer is fit on the whole dataset separately from the input df to work with the decorators
        columns = ["b"]
        df_train = d.create_df_7(library=library)
        df_train = nw.from_native(df_train)

        df_test = df_train.clone()
        expected = self.create_OneHotEncoderTransformer_test_df_1(library=library)

        transformer = OneHotEncodingTransformer(columns=columns)
        transformer.fit(df_train)

        df_transformed = transformer.transform(df_test.to_native())

        expected = nw.from_native(expected)
        for col in [
            column + f"_{value}"
            for column in columns
            for value in df_train.select(nw.col(column).unique())
            .get_column(column)
            .to_list()
        ]:
            expected = expected.with_columns(nw.col(col).cast(nw.Boolean))

        assert_frame_equal_dispatch(expected.to_native(), df_transformed)

        # also test single row transform
        for i in range(len(df_test)):
            df_transformed_row = transformer.transform(df_test[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_categories_not_modified(self, library):
        """Test that the categories from fit are not changed in transform."""
        df_train = d.create_df_1(library=library)
        df_test = d.create_df_7(library=library)

        transformer = OneHotEncodingTransformer(columns=["a", "b"], verbose=False)
        transformer2 = OneHotEncodingTransformer(columns=["a", "b"], verbose=False)

        transformer.fit(df_train)
        transformer2.fit(df_train)

        transformer.transform(df_test)

        assert (
            transformer2.categories_ == transformer.categories_
        ), f"categories_ modified during transform, pre transform had {transformer2.categories_} but post transform has {transformer.categories_}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_renaming_feature_works_as_expected(self, library):
        """Test OneHotEncodingTransformer.transform() is renaming features correctly."""
        df = d.create_df_7(library=library)
        df = df[["b", "c"]]

        transformer = OneHotEncodingTransformer(
            columns=["b", "c"],
            separator="|",
            drop_original=True,
        )

        transformer.fit(df)

        df_transformed = transformer.transform(df)

        expected_columns = ["b|x", "b|y", "b|z", "c|a", "c|b", "c|c"]

        df_transformed = nw.from_native(df_transformed)
        actual_columns = df_transformed.columns

        assert (
            set(expected_columns) == set(actual_columns)
        ), f"renaming columns feature in OneHotEncodingTransformer.transform, expected {expected_columns} but got {actual_columns}"

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_warning_generated_by_unseen_categories(self, library):
        """Test OneHotEncodingTransformer.transform triggers a warning for unseen categories."""
        df_train = d.create_df_7(library=library)
        df_test = d.create_df_8(library=library)

        transformer = OneHotEncodingTransformer(columns=["a", "b", "c"], verbose=True)

        transformer.fit(df_train)

        with pytest.warns(UserWarning, match="unseen categories"):
            transformer.transform(df_test)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_transform_missing_levels_warning(self, library):
        """Test OneHotEncodingTransformer.transform triggers a warning for missing levels."""
        df_train = d.create_df_7(library=library)
        df_test = d.create_df_8(library=library)

        transformer = OneHotEncodingTransformer(
            columns=["b"],
            wanted_values={"b": ["v", "x", "z"]},
        )

        transformer.fit(df_train)

        with pytest.warns(
            UserWarning,
            match=r"OneHotEncodingTransformer: column b includes user-specified values \['v'\] not found in the dataset",
        ):
            transformer.transform(df_test)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_unseen_categories_encoded_as_all_zeroes(self, library):
        """Test OneHotEncodingTransformer.transform encodes unseen categories correctly (all 0s)."""
        # transformer is fit on the whole dataset separately from the input df to work with the decorators
        df_train = d.create_df_7(library=library)

        columns = ["a", "b", "c"]
        x = OneHotEncodingTransformer(columns=columns, verbose=False)
        x.fit(df_train)

        df_test = d.create_df_8(library=library)
        expected = self.create_OneHotEncoderTransformer_test_df_2(library=library)

        df_transformed = x.transform(df_test)

        df_train = nw.from_native(df_train)
        expected = nw.from_native(expected)

        for col in [
            column + f"_{value}"
            for column in columns
            for value in df_train.select(nw.col(column).unique())
            .get_column(column)
            .to_list()
        ]:
            expected = expected.with_columns(nw.col(col).cast(nw.Boolean))

        column_order = expected.columns
        assert_frame_equal_dispatch(expected.to_native(), df_transformed[column_order])

        # also test single row transform
        df_test = nw.from_native(df_test)
        for i in range(len(df_test)):
            df_transformed_row = x.transform(df_test[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            assert_frame_equal_dispatch(
                df_transformed_row[column_order],
                df_expected_row,
            )

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_transform_output_with_wanted_values_arg(self, library):
        """
        Test to verify OneHotEncodingTransformer.transform zero-filled levels from user-specified "wanted_levels" and encodes only those listed in "wanted_levels".

        """
        df_train = d.create_df_7(library=library)
        df_test = d.create_df_8(library=library)

        transformer = OneHotEncodingTransformer(
            columns=["b"],
            wanted_values={"b": ["v", "x", "z"]},
        )

        transformer.fit(df_train)
        df_transformed = transformer.transform(df_test)

        expected_df_dict = {
            "a": [1, 5, 2, 3, 3],
            "b": ["w", "w", "z", "y", "x"],
            "c": ["a", "a", "c", "b", "a"],
            "b_v": [0] * 5,
            "b_x": [0, 0, 0, 0, 1],
            "b_z": [0, 0, 1, 0, 0],
        }
        expected_df = dataframe_init_dispatch(
            library=library,
            dataframe_dict=expected_df_dict,
        )
        expected_df = nw.from_native(expected_df)
        # cast the columns
        boolean_cols = ["b_v", "b_x", "b_z"]
        for col_name in boolean_cols:
            expected_df = expected_df.with_columns(
                nw.col(col_name).cast(nw.Boolean),
            )
        expected_df = expected_df.with_columns(
            nw.col("c").cast(nw.Categorical),
        )

        assert_frame_equal_dispatch(df_transformed, expected_df.to_native())

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_transform_no_warning_if_all_wanted_values_present(self, library, recwarn):
        """Test that OneHotEncodingTransformer.transform does NOT raise a warning when all levels in wanted_levels are present in the data."""
        df_train = d.create_df_8(library=library)
        df_test = d.create_df_7(library=library)

        transformer = OneHotEncodingTransformer(
            columns=["b"],
            wanted_values={"b": ["z", "y", "x"]},
        )
        transformer.fit(df_train)
        transformer.transform(df_test)

        assert (
            len(recwarn) == 0
        ), "OneHotEncodingTransformer.transform is raising unexpected warnings"



import pytest

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
    WeightColumnFitMixinTests,
    WeightColumnInitMixinTests,
)
from tests.imputers.test_BaseImputer import (
    GenericImputerTransformTests,
    GenericImputerTransformTestsWeight,
)
from tubular.imputers import MeanImputer


class TestInit(WeightColumnInitMixinTests, ColumnStrListInitTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanImputer"


class TestFit(WeightColumnFitMixinTests, GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanImputer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values(self, library):
        """Test that the impute values learnt during fit are expected."""
        df = d.create_df_3(library=library)

        x = MeanImputer(columns=["a", "b", "c"])

        x.fit(df)

        expected_impute_values = {
            "a": df["a"].mean(),
            "b": df["b"].mean(),
            "c": df["c"].mean(),
        }

        assert (
            x.impute_values_ == expected_impute_values
        ), f"impute_values_attr not as expected, expected {expected_impute_values} but got {x.impute_values_}"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values_weighted(self, library):
        """Test that the impute values learnt during fit are expected - when weights are used."""
        df = d.create_df_9(library=library)

        x = MeanImputer(columns=["a", "b"], weights_column="c")

        x.fit(df)

        expected_impute_values = {
            "a": (3 + 4 + 16 + 36) / (3 + 2 + 4 + 6),
            "b": (10 + 4 + 12 + 10 + 6) / (2 + 1 + 4 + 5 + 6),
        }

        assert (
            x.impute_values_ == expected_impute_values
        ), f"learnt impute_values_ attr not as expected, expected {expected_impute_values} but got {x.impute_values_}"


class TestTransform(
    GenericTransformTests,
    GenericImputerTransformTestsWeight,
    GenericImputerTransformTests,
):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanImputer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour behaviour outside the three standard methods.
    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanImputer"
